import numpy as np
import numba
import scipy.optimize as sopt
import json

sin = np.sin
cos = np.cos
atan2 = np.arctan2
sqrt = np.sqrt 


class iso_milano_ex8p1_4ord_avr_gov_class: 

    def __init__(self): 

        self.t_end = 10.000000 
        self.Dt = 0.0010000 
        self.decimation = 10.000000 
        self.itol = 1e-6 
        self.Dt_max = 0.001000 
        self.Dt_min = 0.001000 
        self.solvern = 5 
        self.imax = 100 
        self.N_x = 7
        self.N_y = 10 
        self.N_z = 5 
        self.N_store = 10000 
        self.params_list = ['X_d', 'X1d', 'T1d0', 'X_q', 'X1q', 'T1q0', 'R_a', 'X_l', 'H', 'D', 'S_n', 'Omega_b', 'omega_s', 'K_a', 'T_r', 'v_pss', 'Droop', 'T_m', 'K_sec', 'K_delta'] 
        self.params_values_list  = [1.81, 0.3, 8.0, 1.76, 0.65, 1.0, 0.003, 0.05, 6.0, 1.0, 1.0, 314.1592653589793, 1.0, 100, 0.1, 0.0, 0.05, 5.0, 0.001, 0.01] 
        self.inputs_ini_list = ['P_l', 'Q_l', 'P_cig', 'Q_cig', 'v_ref'] 
        self.inputs_ini_values_list  = [0.8, 0.2, 0.1, 0.0, 1.0] 
        self.inputs_run_list = ['P_l', 'Q_l', 'P_cig', 'Q_cig', 'v_ref'] 
        self.inputs_run_values_list = [0.8, 0.2, 0.1, 0.0, 1.0] 
        self.outputs_list = ['p_m', 'P_l', 'Q_l', 'P_cig', 'Q_cig'] 
        self.x_list = ['delta', 'omega', 'e1q', 'e1d', 'v_c', 'p_m', 'xi_m'] 
        self.y_run_list = ['i_d', 'i_q', 'v_1', 'theta_1', 'P_t', 'Q_t', 'v_f', 'p_m_ref', 'v_0', 'theta_0'] 
        self.xy_list = self.x_list + self.y_run_list 
        self.y_ini_list = ['i_d', 'i_q', 'v_1', 'theta_1', 'P_t', 'Q_t', 'v_f', 'p_m_ref', 'v_0', 'theta_0'] 
        self.xy_ini_list = self.x_list + self.y_ini_list 
        self.t = 0.0
        self.it = 0
        self.it_store = 0
        self.xy_prev = np.zeros((self.N_x+self.N_y,1))
        self.initialization_tol = 1e-6
        self.N_u = len(self.inputs_run_list) 
        self.sopt_root_method='hybr'
        self.sopt_root_jac=True
        self.u_ini_list = self.inputs_ini_list
        self.u_ini_values_list = self.inputs_ini_values_list
        self.u_run_list = self.inputs_run_list
        self.u_run_values_list = self.inputs_run_values_list
        self.N_u = len(self.u_run_list)
        self.update() 


    def update(self): 

        self.N_steps = int(np.ceil(self.t_end/self.Dt)) 
        dt = [  
              ('t_end', np.float64),
              ('Dt', np.float64),
              ('decimation', np.float64),
              ('itol', np.float64),
              ('Dt_max', np.float64),
              ('Dt_min', np.float64),
              ('solvern', np.int64),
              ('imax', np.int64),
              ('N_steps', np.int64),
              ('N_store', np.int64),
              ('N_x', np.int64),
              ('N_y', np.int64),
              ('N_z', np.int64),
              ('t', np.float64),
              ('it', np.int64),
              ('it_store', np.int64),
              ('idx', np.int64),
              ('idy', np.int64),
              ('f', np.float64, (self.N_x,1)),
              ('x', np.float64, (self.N_x,1)),
              ('x_0', np.float64, (self.N_x,1)),
              ('g', np.float64, (self.N_y,1)),
              ('y_run', np.float64, (self.N_y,1)),
              ('y_ini', np.float64, (self.N_y,1)),
              ('u_run', np.float64, (self.N_u,1)),
              ('y_0', np.float64, (self.N_y,1)),
              ('h', np.float64, (self.N_z,1)),
              ('Fx', np.float64, (self.N_x,self.N_x)),
              ('Fy', np.float64, (self.N_x,self.N_y)),
              ('Gx', np.float64, (self.N_y,self.N_x)),
              ('Gy', np.float64, (self.N_y,self.N_y)),
              ('Fu', np.float64, (self.N_x,self.N_u)),
              ('Gu', np.float64, (self.N_y,self.N_u)),
              ('Hx', np.float64, (self.N_z,self.N_x)),
              ('Hy', np.float64, (self.N_z,self.N_y)),
              ('Hu', np.float64, (self.N_z,self.N_u)),
              ('Fx_ini', np.float64, (self.N_x,self.N_x)),
              ('Fy_ini', np.float64, (self.N_x,self.N_y)),
              ('Gx_ini', np.float64, (self.N_y,self.N_x)),
              ('Gy_ini', np.float64, (self.N_y,self.N_y)),
              ('T', np.float64, (self.N_store+1,1)),
              ('X', np.float64, (self.N_store+1,self.N_x)),
              ('Y', np.float64, (self.N_store+1,self.N_y)),
              ('Z', np.float64, (self.N_store+1,self.N_z)),
              ('iters', np.float64, (self.N_store+1,1)),
             ]

        values = [
                self.t_end,                          
                self.Dt,
                self.decimation,
                self.itol,
                self.Dt_max,
                self.Dt_min,
                self.solvern,
                self.imax,
                self.N_steps,
                self.N_store,
                self.N_x,
                self.N_y,
                self.N_z,
                self.t,
                self.it,
                self.it_store,
                0,                                     # idx
                0,                                     # idy
                np.zeros((self.N_x,1)),                # f
                np.zeros((self.N_x,1)),                # x
                np.zeros((self.N_x,1)),                # x_0
                np.zeros((self.N_y,1)),                # g
                np.zeros((self.N_y,1)),                # y_run
                np.zeros((self.N_y,1)),                # y_ini
                np.zeros((self.N_u,1)),                # u_run
                np.zeros((self.N_y,1)),                # y_0
                np.zeros((self.N_z,1)),                # h
                np.zeros((self.N_x,self.N_x)),         # Fx   
                np.zeros((self.N_x,self.N_y)),         # Fy 
                np.zeros((self.N_y,self.N_x)),         # Gx 
                np.zeros((self.N_y,self.N_y)),         # Fy
                np.zeros((self.N_x,self.N_u)),         # Fu 
                np.zeros((self.N_y,self.N_u)),         # Gu 
                np.zeros((self.N_z,self.N_x)),         # Hx 
                np.zeros((self.N_z,self.N_y)),         # Hy 
                np.zeros((self.N_z,self.N_u)),         # Hu 
                np.zeros((self.N_x,self.N_x)),         # Fx_ini  
                np.zeros((self.N_x,self.N_y)),         # Fy_ini 
                np.zeros((self.N_y,self.N_x)),         # Gx_ini 
                np.zeros((self.N_y,self.N_y)),         # Fy_ini 
                np.zeros((self.N_store+1,1)),          # T
                np.zeros((self.N_store+1,self.N_x)),   # X
                np.zeros((self.N_store+1,self.N_y)),   # Y
                np.zeros((self.N_store+1,self.N_z)),   # Z
                np.zeros((self.N_store+1,1)),          # iters
                ]  

        dt += [(item,np.float64) for item in self.params_list]
        values += [item for item in self.params_values_list]

        for item_id,item_val in zip(self.inputs_ini_list,self.inputs_ini_values_list):
            if item_id in self.inputs_run_list: continue
            dt += [(item_id,np.float64)]
            values += [item_val]

        dt += [(item,np.float64) for item in self.inputs_run_list]
        values += [item for item in self.inputs_run_values_list]

        self.struct = np.rec.array([tuple(values)], dtype=np.dtype(dt))
        
        xy0 = np.zeros((self.N_x+self.N_y,))
        self.ini_dae_jacobian_nn(xy0)
        self.run_dae_jacobian_nn(xy0)

    def load_params(self,data_input):

        if type(data_input) == str:
            json_file = data_input
            self.json_file = json_file
            self.json_data = open(json_file).read().replace("'",'"')
            data = json.loads(self.json_data)
        elif type(data_input) == dict:
            data = data_input

        self.data = data
        for item in self.data:
            self.struct[0][item] = self.data[item]
            self.params_values_list[self.params_list.index(item)] = self.data[item]



    def ini_problem(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        ini(self.struct,2)
        ini(self.struct,3)       
        fg = np.vstack((self.struct[0].f,self.struct[0].g))[:,0]
        return fg

    def run_problem(self,x):
        t = self.struct[0].t
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_run[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        run(t,self.struct,2)
        run(t,self.struct,3)
        run(t,self.struct,10)
        run(t,self.struct,11)
        run(t,self.struct,12)
        run(t,self.struct,13)
        
        fg = np.vstack((self.struct[0].f,self.struct[0].g))[:,0]
        return fg
    

    def run_dae_jacobian(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_run[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        run(0.0,self.struct,10)
        run(0.0,self.struct,11)     
        run(0.0,self.struct,12)
        run(0.0,self.struct,13)
        A_c = np.block([[self.struct[0].Fx,self.struct[0].Fy],
                        [self.struct[0].Gx,self.struct[0].Gy]])
        return A_c

    def run_dae_jacobian_nn(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_run[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        run_nn(0.0,self.struct,10)
        run_nn(0.0,self.struct,11)     
        run_nn(0.0,self.struct,12)
        run_nn(0.0,self.struct,13)
 

    
    def eval_jacobians(self):

        run(0.0,self.struct,10)
        run(0.0,self.struct,11)  
        run(0.0,self.struct,12) 

        return 1


    def ini_dae_jacobian(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        ini(self.struct,10)
        ini(self.struct,11)       
        A_c = np.block([[self.struct[0].Fx_ini,self.struct[0].Fy_ini],
                        [self.struct[0].Gx_ini,self.struct[0].Gy_ini]])
        return A_c

    def ini_dae_jacobian_nn(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        ini_nn(self.struct,10)
        ini_nn(self.struct,11)       
 

    def f_ode(self,x):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def f_odeint(self,x,t):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def f_ivp(self,t,x):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def Fx_ode(self,x):
        self.struct[0].x[:,0] = x
        run(self.struct,10)
        return self.struct[0].Fx

    def eval_A(self):
        
        Fx = self.struct[0].Fx
        Fy = self.struct[0].Fy
        Gx = self.struct[0].Gx
        Gy = self.struct[0].Gy
        
        A = Fx - Fy @ np.linalg.solve(Gy,Gx)
        
        self.A = A
        
        return A

    def eval_A_ini(self):
        
        Fx = self.struct[0].Fx_ini
        Fy = self.struct[0].Fy_ini
        Gx = self.struct[0].Gx_ini
        Gy = self.struct[0].Gy_ini
        
        A = Fx - Fy @ np.linalg.solve(Gy,Gx)
        
        
        return A
    
    def reset(self):
        for param,param_value in zip(self.params_list,self.params_values_list):
            self.struct[0][param] = param_value
        for input_name,input_value in zip(self.inputs_ini_list,self.inputs_ini_values_list):
            self.struct[0][input_name] = input_value   
        for input_name,input_value in zip(self.inputs_run_list,self.inputs_run_values_list):
            self.struct[0][input_name] = input_value  

    def simulate(self,events,xy0=0):
        
        # initialize both the ini and the run system
        self.initialize(events,xy0=xy0)
        
        # simulation run
        for event in events:  
            # make all the desired changes
            self.run([event]) 
            
        # post process
        T,X,Y,Z = self.post()
        
        return T,X,Y,Z
    

    
    def run(self,events):
        

        # simulation run
        for event in events:  
            # make all the desired changes
            for item in event:
                self.struct[0][item] = event[item]
            daesolver(self.struct)    # run until next event
            
        return 1
    
    
    def post(self):
        
        # post process result    
        T = self.struct[0]['T'][:self.struct[0].it_store]
        X = self.struct[0]['X'][:self.struct[0].it_store,:]
        Y = self.struct[0]['Y'][:self.struct[0].it_store,:]
        Z = self.struct[0]['Z'][:self.struct[0].it_store,:]
        iters = self.struct[0]['iters'][:self.struct[0].it_store,:]
    
        self.T = T
        self.X = X
        self.Y = Y
        self.Z = Z
        self.iters = iters
        
        return T,X,Y,Z
        
    def save_0(self,file_name = 'xy_0.json'):
        xy_0_dict = {}
        for item in self.x_list:
            xy_0_dict.update({item:self.get_value(item)})
        for item in self.y_ini_list:
            xy_0_dict.update({item:self.get_value(item)})
    
        xy_0_str = json.dumps(xy_0_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(xy_0_str)

    def load_0(self,file_name = 'xy_0.json'):
        with open(file_name) as fobj:
            xy_0_str = fobj.read()
        xy_0_dict = json.loads(xy_0_str)
    
        for item in xy_0_dict:
            if item in self.x_list:
                self.xy_prev[self.x_list.index(item)] = xy_0_dict[item]
            if item in self.y_ini_list:
                self.xy_prev[self.y_ini_list.index(item)+self.N_x] = xy_0_dict[item]
                
            
    def initialize(self,events=[{}],xy0=0):
        '''
        

        Parameters
        ----------
        events : dictionary 
            Dictionary with at least 't_end' and all inputs and parameters 
            that need to be changed.
        xy0 : float or string, optional
            0 means all states should be zero as initial guess. 
            If not zero all the states initial guess are the given input.
            If 'prev' it uses the last known initialization result as initial guess.

        Returns
        -------
        T : TYPE
            DESCRIPTION.
        X : TYPE
            DESCRIPTION.
        Y : TYPE
            DESCRIPTION.
        Z : TYPE
            DESCRIPTION.

        '''
        # simulation parameters
        self.struct[0].it = 0       # set time step to zero
        self.struct[0].it_store = 0 # set storage to zero
        self.struct[0].t = 0.0      # set time to zero
                    
        # initialization
        it_event = 0
        event = events[it_event]
        for item in event:
            self.struct[0][item] = event[item]
            
        
        ## compute initial conditions using x and y_ini 
        if type(xy0) == str:
            if xy0 == 'prev':
                xy0 = self.xy_prev
            else:
                self.load_0(xy0)
                xy0 = self.xy_prev
        else:
            if xy0 == 0:
                xy0 = np.zeros(self.N_x+self.N_y)
            elif xy0 == 1:
                xy0 = np.ones(self.N_x+self.N_y)
            else:
                xy0 = xy0*np.ones(self.N_x+self.N_y)

        #xy = sopt.fsolve(self.ini_problem,xy0, jac=self.ini_dae_jacobian )

        
        if self.sopt_root_jac:
            sol = sopt.root(self.ini_problem, xy0, 
                            jac=self.ini_dae_jacobian, 
                            method=self.sopt_root_method, tol=self.initialization_tol)
        else:
            sol = sopt.root(self.ini_problem, xy0, method=self.sopt_root_method)

        self.initialization_ok = True
        if sol.success == False:
            print('initialization not found!')
            self.initialization_ok = False

            T = self.struct[0]['T'][:self.struct[0].it_store]
            X = self.struct[0]['X'][:self.struct[0].it_store,:]
            Y = self.struct[0]['Y'][:self.struct[0].it_store,:]
            Z = self.struct[0]['Z'][:self.struct[0].it_store,:]
            iters = self.struct[0]['iters'][:self.struct[0].it_store,:]

        if self.initialization_ok:
            xy = sol.x
            self.xy_prev = xy
            self.struct[0].x[:,0] = xy[0:self.N_x]
            self.struct[0].y_run[:,0] = xy[self.N_x:]

            ## y_ini to u_run
            for item in self.inputs_run_list:
                if item in self.y_ini_list:
                    self.struct[0][item] = self.struct[0].y_ini[self.y_ini_list.index(item)]

            ## u_ini to y_run
            for item in self.inputs_ini_list:
                if item in self.y_run_list:
                    self.struct[0].y_run[self.y_run_list.index(item)] = self.struct[0][item]


            #xy = sopt.fsolve(self.ini_problem,xy0, jac=self.ini_dae_jacobian )
            if self.sopt_root_jac:
                sol = sopt.root(self.run_problem, xy0, 
                                jac=self.run_dae_jacobian, 
                                method=self.sopt_root_method, tol=self.initialization_tol)
            else:
                sol = sopt.root(self.run_problem, xy0, method=self.sopt_root_method)

            # evaluate f and g
            run(0.0,self.struct,2)
            run(0.0,self.struct,3)                

            
            # evaluate run jacobians 
            run(0.0,self.struct,10)
            run(0.0,self.struct,11)                
            run(0.0,self.struct,12) 
            run(0.0,self.struct,14) 
             
            # post process result    
            T = self.struct[0]['T'][:self.struct[0].it_store]
            X = self.struct[0]['X'][:self.struct[0].it_store,:]
            Y = self.struct[0]['Y'][:self.struct[0].it_store,:]
            Z = self.struct[0]['Z'][:self.struct[0].it_store,:]
            iters = self.struct[0]['iters'][:self.struct[0].it_store,:]
        
            self.T = T
            self.X = X
            self.Y = Y
            self.Z = Z
            self.iters = iters
            
        return self.initialization_ok
    
    
    def get_value(self,name):
        if name in self.inputs_run_list:
            value = self.struct[0][name]
        if name in self.x_list:
            idx = self.x_list.index(name)
            value = self.struct[0].x[idx,0]
        if name in self.y_run_list:
            idy = self.y_run_list.index(name)
            value = self.struct[0].y_run[idy,0]
        if name in self.params_list:
            value = self.struct[0][name]
        if name in self.outputs_list:
            value = self.struct[0].h[self.outputs_list.index(name),0] 

        return value
    
    def get_values(self,name):
        if name in self.x_list:
            values = self.X[:,self.x_list.index(name)]
        if name in self.y_run_list:
            values = self.Y[:,self.y_run_list.index(name)]
        if name in self.outputs_list:
            values = self.Z[:,self.outputs_list.index(name)]
                        
        return values

    def get_mvalue(self,names):
        '''

        Parameters
        ----------
        names : list
            list of variables names to return each value.

        Returns
        -------
        mvalue : TYPE
            list of value of each variable.

        '''
        mvalue = []
        for name in names:
            mvalue += [self.get_value(name)]
                        
        return mvalue
    
    def set_value(self,name,value):
        if name in self.inputs_run_list:
            self.struct[0][name] = value
        if name in self.params_list:
            self.struct[0][name] = value
            
    def report_x(self,value_format='5.2f'):
        for item in self.x_list:
            print(f'{item:5s} = {self.get_value(item):5.2f}')

    def report_y(self,value_format='5.2f'):
        for item in self.y_run_list:
            print(f'{item:5s} = {self.get_value(item):5.2f}')
            
    def report_u(self,value_format='5.2f'):
        for item in self.inputs_run_list:
            print(f'{item:5s} = {self.get_value(item):5.2f}')

    def report_z(self,value_format='5.2f'):
        for item in self.outputs_list:
            print(f'{item:5s} = {self.get_value(item):5.2f}')
            
    def get_x(self):
        return self.struct[0].x


@numba.njit(cache=True)
def ini(struct,mode):

    # Parameters:
    X_d = struct[0].X_d
    X1d = struct[0].X1d
    T1d0 = struct[0].T1d0
    X_q = struct[0].X_q
    X1q = struct[0].X1q
    T1q0 = struct[0].T1q0
    R_a = struct[0].R_a
    X_l = struct[0].X_l
    H = struct[0].H
    D = struct[0].D
    S_n = struct[0].S_n
    Omega_b = struct[0].Omega_b
    omega_s = struct[0].omega_s
    K_a = struct[0].K_a
    T_r = struct[0].T_r
    v_pss = struct[0].v_pss
    Droop = struct[0].Droop
    T_m = struct[0].T_m
    K_sec = struct[0].K_sec
    K_delta = struct[0].K_delta
    
    # Inputs:
    P_l = struct[0].P_l
    Q_l = struct[0].Q_l
    P_cig = struct[0].P_cig
    Q_cig = struct[0].Q_cig
    v_ref = struct[0].v_ref
    
    # Dynamical states:
    delta = struct[0].x[0,0]
    omega = struct[0].x[1,0]
    e1q = struct[0].x[2,0]
    e1d = struct[0].x[3,0]
    v_c = struct[0].x[4,0]
    p_m = struct[0].x[5,0]
    xi_m = struct[0].x[6,0]
    
    # Algebraic states:
    i_d = struct[0].y_ini[0,0]
    i_q = struct[0].y_ini[1,0]
    v_1 = struct[0].y_ini[2,0]
    theta_1 = struct[0].y_ini[3,0]
    P_t = struct[0].y_ini[4,0]
    Q_t = struct[0].y_ini[5,0]
    v_f = struct[0].y_ini[6,0]
    p_m_ref = struct[0].y_ini[7,0]
    v_0 = struct[0].y_ini[8,0]
    theta_0 = struct[0].y_ini[9,0]
    
    # Differential equations:
    if mode == 2:


        struct[0].f[0,0] = -K_delta*delta + Omega_b*(omega - omega_s)
        struct[0].f[1,0] = (-D*(omega - omega_s) - i_d*(R_a*i_d + v_1*sin(delta - theta_1)) - i_q*(R_a*i_q + v_1*cos(delta - theta_1)) + p_m)/(2*H)
        struct[0].f[2,0] = (-e1q - i_d*(-X1d + X_d) + v_f)/T1d0
        struct[0].f[3,0] = (-e1d + i_q*(-X1q + X_q))/T1q0
        struct[0].f[4,0] = (v_1 - v_c)/T_r
        struct[0].f[5,0] = (-p_m + p_m_ref)/T_m
        struct[0].f[6,0] = omega - 1
    
    # Algebraic equations:
    if mode == 3:

        g_n = np.ascontiguousarray(struct[0].Gy_ini) @ np.ascontiguousarray(struct[0].y_ini)

        struct[0].g[0,0] = R_a*i_q + X1d*i_d - e1q + v_1*cos(delta - theta_1)
        struct[0].g[1,0] = R_a*i_d - X1q*i_q - e1d + v_1*sin(delta - theta_1)
        struct[0].g[2,0] = P_cig + P_t*S_n + v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g[3,0] = Q_cig + Q_t*S_n + v_0*v_1*cos(theta_0 - theta_1)/X_l - v_1**2/X_l
        struct[0].g[4,0] = -P_t + i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].g[5,0] = -Q_t + i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].g[6,0] = K_a*(-v_c + v_pss + v_ref) - v_f
        struct[0].g[7,0] = -K_sec*xi_m - p_m_ref - (omega - 1)/Droop
        struct[0].g[8,0] = -P_l - v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g[9,0] = -Q_l - v_0**2/X_l + v_0*v_1*cos(theta_0 - theta_1)/X_l
    
    # Outputs:
    if mode == 3:

        struct[0].h[0,0] = p_m
        struct[0].h[1,0] = P_l
        struct[0].h[2,0] = Q_l
        struct[0].h[3,0] = P_cig
        struct[0].h[4,0] = Q_cig
    

    if mode == 10:

        struct[0].Fx_ini[0,0] = -K_delta
        struct[0].Fx_ini[0,1] = Omega_b
        struct[0].Fx_ini[1,0] = (-i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fx_ini[1,1] = -D/(2*H)
        struct[0].Fx_ini[1,5] = 1/(2*H)
        struct[0].Fx_ini[2,2] = -1/T1d0
        struct[0].Fx_ini[3,3] = -1/T1q0
        struct[0].Fx_ini[4,4] = -1/T_r
        struct[0].Fx_ini[5,5] = -1/T_m

    if mode == 11:

        struct[0].Fy_ini[1,0] = (-2*R_a*i_d - v_1*sin(delta - theta_1))/(2*H) 
        struct[0].Fy_ini[1,1] = (-2*R_a*i_q - v_1*cos(delta - theta_1))/(2*H) 
        struct[0].Fy_ini[1,2] = (-i_d*sin(delta - theta_1) - i_q*cos(delta - theta_1))/(2*H) 
        struct[0].Fy_ini[1,3] = (i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1))/(2*H) 
        struct[0].Fy_ini[2,0] = (X1d - X_d)/T1d0 
        struct[0].Fy_ini[2,6] = 1/T1d0 
        struct[0].Fy_ini[3,1] = (-X1q + X_q)/T1q0 
        struct[0].Fy_ini[4,2] = 1/T_r 
        struct[0].Fy_ini[5,7] = 1/T_m 

        struct[0].Gx_ini[0,0] = -v_1*sin(delta - theta_1)
        struct[0].Gx_ini[0,2] = -1
        struct[0].Gx_ini[1,0] = v_1*cos(delta - theta_1)
        struct[0].Gx_ini[1,3] = -1
        struct[0].Gx_ini[4,0] = i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].Gx_ini[5,0] = -i_d*v_1*sin(delta - theta_1) - i_q*v_1*cos(delta - theta_1)
        struct[0].Gx_ini[6,4] = -K_a
        struct[0].Gx_ini[7,1] = -1/Droop
        struct[0].Gx_ini[7,6] = -K_sec

        struct[0].Gy_ini[0,0] = X1d
        struct[0].Gy_ini[0,1] = R_a
        struct[0].Gy_ini[0,2] = cos(delta - theta_1)
        struct[0].Gy_ini[0,3] = v_1*sin(delta - theta_1)
        struct[0].Gy_ini[1,0] = R_a
        struct[0].Gy_ini[1,1] = -X1q
        struct[0].Gy_ini[1,2] = sin(delta - theta_1)
        struct[0].Gy_ini[1,3] = -v_1*cos(delta - theta_1)
        struct[0].Gy_ini[2,2] = v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[2,3] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[2,4] = S_n
        struct[0].Gy_ini[2,8] = v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[2,9] = v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[3,2] = v_0*cos(theta_0 - theta_1)/X_l - 2*v_1/X_l
        struct[0].Gy_ini[3,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[3,5] = S_n
        struct[0].Gy_ini[3,8] = v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[3,9] = -v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[4,0] = v_1*sin(delta - theta_1)
        struct[0].Gy_ini[4,1] = v_1*cos(delta - theta_1)
        struct[0].Gy_ini[4,2] = i_d*sin(delta - theta_1) + i_q*cos(delta - theta_1)
        struct[0].Gy_ini[4,3] = -i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1)
        struct[0].Gy_ini[5,0] = v_1*cos(delta - theta_1)
        struct[0].Gy_ini[5,1] = -v_1*sin(delta - theta_1)
        struct[0].Gy_ini[5,2] = i_d*cos(delta - theta_1) - i_q*sin(delta - theta_1)
        struct[0].Gy_ini[5,3] = i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].Gy_ini[8,2] = -v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[8,3] = v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[8,8] = -v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[8,9] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[9,2] = v_0*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[9,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[9,8] = -2*v_0/X_l + v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[9,9] = -v_0*v_1*sin(theta_0 - theta_1)/X_l



@numba.njit(cache=True)
def run(t,struct,mode):

    # Parameters:
    X_d = struct[0].X_d
    X1d = struct[0].X1d
    T1d0 = struct[0].T1d0
    X_q = struct[0].X_q
    X1q = struct[0].X1q
    T1q0 = struct[0].T1q0
    R_a = struct[0].R_a
    X_l = struct[0].X_l
    H = struct[0].H
    D = struct[0].D
    S_n = struct[0].S_n
    Omega_b = struct[0].Omega_b
    omega_s = struct[0].omega_s
    K_a = struct[0].K_a
    T_r = struct[0].T_r
    v_pss = struct[0].v_pss
    Droop = struct[0].Droop
    T_m = struct[0].T_m
    K_sec = struct[0].K_sec
    K_delta = struct[0].K_delta
    
    # Inputs:
    P_l = struct[0].P_l
    Q_l = struct[0].Q_l
    P_cig = struct[0].P_cig
    Q_cig = struct[0].Q_cig
    v_ref = struct[0].v_ref
    
    # Dynamical states:
    delta = struct[0].x[0,0]
    omega = struct[0].x[1,0]
    e1q = struct[0].x[2,0]
    e1d = struct[0].x[3,0]
    v_c = struct[0].x[4,0]
    p_m = struct[0].x[5,0]
    xi_m = struct[0].x[6,0]
    
    # Algebraic states:
    i_d = struct[0].y_run[0,0]
    i_q = struct[0].y_run[1,0]
    v_1 = struct[0].y_run[2,0]
    theta_1 = struct[0].y_run[3,0]
    P_t = struct[0].y_run[4,0]
    Q_t = struct[0].y_run[5,0]
    v_f = struct[0].y_run[6,0]
    p_m_ref = struct[0].y_run[7,0]
    v_0 = struct[0].y_run[8,0]
    theta_0 = struct[0].y_run[9,0]
    
    struct[0].u_run[0,0] = P_l
    struct[0].u_run[1,0] = Q_l
    struct[0].u_run[2,0] = P_cig
    struct[0].u_run[3,0] = Q_cig
    struct[0].u_run[4,0] = v_ref
    # Differential equations:
    if mode == 2:


        struct[0].f[0,0] = -K_delta*delta + Omega_b*(omega - omega_s)
        struct[0].f[1,0] = (-D*(omega - omega_s) - i_d*(R_a*i_d + v_1*sin(delta - theta_1)) - i_q*(R_a*i_q + v_1*cos(delta - theta_1)) + p_m)/(2*H)
        struct[0].f[2,0] = (-e1q - i_d*(-X1d + X_d) + v_f)/T1d0
        struct[0].f[3,0] = (-e1d + i_q*(-X1q + X_q))/T1q0
        struct[0].f[4,0] = (v_1 - v_c)/T_r
        struct[0].f[5,0] = (-p_m + p_m_ref)/T_m
        struct[0].f[6,0] = omega - 1
    
    # Algebraic equations:
    if mode == 3:

        g_n = np.ascontiguousarray(struct[0].Gy) @ np.ascontiguousarray(struct[0].y_run) + np.ascontiguousarray(struct[0].Gu) @ np.ascontiguousarray(struct[0].u_run)

        struct[0].g[0,0] = R_a*i_q + X1d*i_d - e1q + v_1*cos(delta - theta_1)
        struct[0].g[1,0] = R_a*i_d - X1q*i_q - e1d + v_1*sin(delta - theta_1)
        struct[0].g[2,0] = P_cig + P_t*S_n + v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g[3,0] = Q_cig + Q_t*S_n + v_0*v_1*cos(theta_0 - theta_1)/X_l - v_1**2/X_l
        struct[0].g[4,0] = -P_t + i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].g[5,0] = -Q_t + i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].g[6,0] = K_a*(-v_c + v_pss + v_ref) - v_f
        struct[0].g[7,0] = -K_sec*xi_m - p_m_ref - (omega - 1)/Droop
        struct[0].g[8,0] = -P_l - v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g[9,0] = -Q_l - v_0**2/X_l + v_0*v_1*cos(theta_0 - theta_1)/X_l
    
    # Outputs:
    if mode == 3:

        struct[0].h[0,0] = p_m
        struct[0].h[1,0] = P_l
        struct[0].h[2,0] = Q_l
        struct[0].h[3,0] = P_cig
        struct[0].h[4,0] = Q_cig
    

    if mode == 10:

        struct[0].Fx[0,0] = -K_delta
        struct[0].Fx[0,1] = Omega_b
        struct[0].Fx[1,0] = (-i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fx[1,1] = -D/(2*H)
        struct[0].Fx[1,5] = 1/(2*H)
        struct[0].Fx[2,2] = -1/T1d0
        struct[0].Fx[3,3] = -1/T1q0
        struct[0].Fx[4,4] = -1/T_r
        struct[0].Fx[5,5] = -1/T_m

    if mode == 11:

        struct[0].Fy[1,0] = (-2*R_a*i_d - v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fy[1,1] = (-2*R_a*i_q - v_1*cos(delta - theta_1))/(2*H)
        struct[0].Fy[1,2] = (-i_d*sin(delta - theta_1) - i_q*cos(delta - theta_1))/(2*H)
        struct[0].Fy[1,3] = (i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fy[2,0] = (X1d - X_d)/T1d0
        struct[0].Fy[2,6] = 1/T1d0
        struct[0].Fy[3,1] = (-X1q + X_q)/T1q0
        struct[0].Fy[4,2] = 1/T_r
        struct[0].Fy[5,7] = 1/T_m

        struct[0].Gx[0,0] = -v_1*sin(delta - theta_1)
        struct[0].Gx[0,2] = -1
        struct[0].Gx[1,0] = v_1*cos(delta - theta_1)
        struct[0].Gx[1,3] = -1
        struct[0].Gx[4,0] = i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].Gx[5,0] = -i_d*v_1*sin(delta - theta_1) - i_q*v_1*cos(delta - theta_1)
        struct[0].Gx[6,4] = -K_a
        struct[0].Gx[7,1] = -1/Droop
        struct[0].Gx[7,6] = -K_sec

        struct[0].Gy[0,0] = X1d
        struct[0].Gy[0,1] = R_a
        struct[0].Gy[0,2] = cos(delta - theta_1)
        struct[0].Gy[0,3] = v_1*sin(delta - theta_1)
        struct[0].Gy[1,0] = R_a
        struct[0].Gy[1,1] = -X1q
        struct[0].Gy[1,2] = sin(delta - theta_1)
        struct[0].Gy[1,3] = -v_1*cos(delta - theta_1)
        struct[0].Gy[2,2] = v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[2,3] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[2,4] = S_n
        struct[0].Gy[2,8] = v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[2,9] = v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[3,2] = v_0*cos(theta_0 - theta_1)/X_l - 2*v_1/X_l
        struct[0].Gy[3,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[3,5] = S_n
        struct[0].Gy[3,8] = v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[3,9] = -v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[4,0] = v_1*sin(delta - theta_1)
        struct[0].Gy[4,1] = v_1*cos(delta - theta_1)
        struct[0].Gy[4,2] = i_d*sin(delta - theta_1) + i_q*cos(delta - theta_1)
        struct[0].Gy[4,3] = -i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1)
        struct[0].Gy[5,0] = v_1*cos(delta - theta_1)
        struct[0].Gy[5,1] = -v_1*sin(delta - theta_1)
        struct[0].Gy[5,2] = i_d*cos(delta - theta_1) - i_q*sin(delta - theta_1)
        struct[0].Gy[5,3] = i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].Gy[8,2] = -v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[8,3] = v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[8,8] = -v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[8,9] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[9,2] = v_0*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[9,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[9,8] = -2*v_0/X_l + v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[9,9] = -v_0*v_1*sin(theta_0 - theta_1)/X_l

    if mode > 12:



        struct[0].Hx[0,5] = 1





def ini_nn(struct,mode):

    # Parameters:
    X_d = struct[0].X_d
    X1d = struct[0].X1d
    T1d0 = struct[0].T1d0
    X_q = struct[0].X_q
    X1q = struct[0].X1q
    T1q0 = struct[0].T1q0
    R_a = struct[0].R_a
    X_l = struct[0].X_l
    H = struct[0].H
    D = struct[0].D
    S_n = struct[0].S_n
    Omega_b = struct[0].Omega_b
    omega_s = struct[0].omega_s
    K_a = struct[0].K_a
    T_r = struct[0].T_r
    v_pss = struct[0].v_pss
    Droop = struct[0].Droop
    T_m = struct[0].T_m
    K_sec = struct[0].K_sec
    K_delta = struct[0].K_delta
    
    # Inputs:
    P_l = struct[0].P_l
    Q_l = struct[0].Q_l
    P_cig = struct[0].P_cig
    Q_cig = struct[0].Q_cig
    v_ref = struct[0].v_ref
    
    # Dynamical states:
    delta = struct[0].x[0,0]
    omega = struct[0].x[1,0]
    e1q = struct[0].x[2,0]
    e1d = struct[0].x[3,0]
    v_c = struct[0].x[4,0]
    p_m = struct[0].x[5,0]
    xi_m = struct[0].x[6,0]
    
    # Algebraic states:
    i_d = struct[0].y_ini[0,0]
    i_q = struct[0].y_ini[1,0]
    v_1 = struct[0].y_ini[2,0]
    theta_1 = struct[0].y_ini[3,0]
    P_t = struct[0].y_ini[4,0]
    Q_t = struct[0].y_ini[5,0]
    v_f = struct[0].y_ini[6,0]
    p_m_ref = struct[0].y_ini[7,0]
    v_0 = struct[0].y_ini[8,0]
    theta_0 = struct[0].y_ini[9,0]
    
    # Differential equations:
    if mode == 2:


        struct[0].f[0,0] = -K_delta*delta + Omega_b*(omega - omega_s)
        struct[0].f[1,0] = (-D*(omega - omega_s) - i_d*(R_a*i_d + v_1*sin(delta - theta_1)) - i_q*(R_a*i_q + v_1*cos(delta - theta_1)) + p_m)/(2*H)
        struct[0].f[2,0] = (-e1q - i_d*(-X1d + X_d) + v_f)/T1d0
        struct[0].f[3,0] = (-e1d + i_q*(-X1q + X_q))/T1q0
        struct[0].f[4,0] = (v_1 - v_c)/T_r
        struct[0].f[5,0] = (-p_m + p_m_ref)/T_m
        struct[0].f[6,0] = omega - 1
    
    # Algebraic equations:
    if mode == 3:


        struct[0].g[0,0] = R_a*i_q + X1d*i_d - e1q + v_1*cos(delta - theta_1)
        struct[0].g[1,0] = R_a*i_d - X1q*i_q - e1d + v_1*sin(delta - theta_1)
        struct[0].g[2,0] = P_cig + P_t*S_n + v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g[3,0] = Q_cig + Q_t*S_n + v_0*v_1*cos(theta_0 - theta_1)/X_l - v_1**2/X_l
        struct[0].g[4,0] = -P_t + i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].g[5,0] = -Q_t + i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].g[6,0] = K_a*(-v_c + v_pss + v_ref) - v_f
        struct[0].g[7,0] = -K_sec*xi_m - p_m_ref - (omega - 1)/Droop
        struct[0].g[8,0] = -P_l - v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g[9,0] = -Q_l - v_0**2/X_l + v_0*v_1*cos(theta_0 - theta_1)/X_l
    
    # Outputs:
    if mode == 3:

        struct[0].h[0,0] = p_m
        struct[0].h[1,0] = P_l
        struct[0].h[2,0] = Q_l
        struct[0].h[3,0] = P_cig
        struct[0].h[4,0] = Q_cig
    

    if mode == 10:

        struct[0].Fx_ini[0,0] = -K_delta
        struct[0].Fx_ini[0,1] = Omega_b
        struct[0].Fx_ini[1,0] = (-i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fx_ini[1,1] = -D/(2*H)
        struct[0].Fx_ini[1,5] = 1/(2*H)
        struct[0].Fx_ini[2,2] = -1/T1d0
        struct[0].Fx_ini[3,3] = -1/T1q0
        struct[0].Fx_ini[4,4] = -1/T_r
        struct[0].Fx_ini[5,5] = -1/T_m
        struct[0].Fx_ini[6,1] = 1

    if mode == 11:

        struct[0].Fy_ini[1,0] = (-2*R_a*i_d - v_1*sin(delta - theta_1))/(2*H) 
        struct[0].Fy_ini[1,1] = (-2*R_a*i_q - v_1*cos(delta - theta_1))/(2*H) 
        struct[0].Fy_ini[1,2] = (-i_d*sin(delta - theta_1) - i_q*cos(delta - theta_1))/(2*H) 
        struct[0].Fy_ini[1,3] = (i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1))/(2*H) 
        struct[0].Fy_ini[2,0] = (X1d - X_d)/T1d0 
        struct[0].Fy_ini[2,6] = 1/T1d0 
        struct[0].Fy_ini[3,1] = (-X1q + X_q)/T1q0 
        struct[0].Fy_ini[4,2] = 1/T_r 
        struct[0].Fy_ini[5,7] = 1/T_m 

        struct[0].Gy_ini[0,0] = X1d
        struct[0].Gy_ini[0,1] = R_a
        struct[0].Gy_ini[0,2] = cos(delta - theta_1)
        struct[0].Gy_ini[0,3] = v_1*sin(delta - theta_1)
        struct[0].Gy_ini[1,0] = R_a
        struct[0].Gy_ini[1,1] = -X1q
        struct[0].Gy_ini[1,2] = sin(delta - theta_1)
        struct[0].Gy_ini[1,3] = -v_1*cos(delta - theta_1)
        struct[0].Gy_ini[2,2] = v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[2,3] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[2,4] = S_n
        struct[0].Gy_ini[2,8] = v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[2,9] = v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[3,2] = v_0*cos(theta_0 - theta_1)/X_l - 2*v_1/X_l
        struct[0].Gy_ini[3,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[3,5] = S_n
        struct[0].Gy_ini[3,8] = v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[3,9] = -v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[4,0] = v_1*sin(delta - theta_1)
        struct[0].Gy_ini[4,1] = v_1*cos(delta - theta_1)
        struct[0].Gy_ini[4,2] = i_d*sin(delta - theta_1) + i_q*cos(delta - theta_1)
        struct[0].Gy_ini[4,3] = -i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1)
        struct[0].Gy_ini[4,4] = -1
        struct[0].Gy_ini[5,0] = v_1*cos(delta - theta_1)
        struct[0].Gy_ini[5,1] = -v_1*sin(delta - theta_1)
        struct[0].Gy_ini[5,2] = i_d*cos(delta - theta_1) - i_q*sin(delta - theta_1)
        struct[0].Gy_ini[5,3] = i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].Gy_ini[5,5] = -1
        struct[0].Gy_ini[6,6] = -1
        struct[0].Gy_ini[7,7] = -1
        struct[0].Gy_ini[8,2] = -v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[8,3] = v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[8,8] = -v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[8,9] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[9,2] = v_0*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[9,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[9,8] = -2*v_0/X_l + v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[9,9] = -v_0*v_1*sin(theta_0 - theta_1)/X_l



def run_nn(t,struct,mode):

    # Parameters:
    X_d = struct[0].X_d
    X1d = struct[0].X1d
    T1d0 = struct[0].T1d0
    X_q = struct[0].X_q
    X1q = struct[0].X1q
    T1q0 = struct[0].T1q0
    R_a = struct[0].R_a
    X_l = struct[0].X_l
    H = struct[0].H
    D = struct[0].D
    S_n = struct[0].S_n
    Omega_b = struct[0].Omega_b
    omega_s = struct[0].omega_s
    K_a = struct[0].K_a
    T_r = struct[0].T_r
    v_pss = struct[0].v_pss
    Droop = struct[0].Droop
    T_m = struct[0].T_m
    K_sec = struct[0].K_sec
    K_delta = struct[0].K_delta
    
    # Inputs:
    P_l = struct[0].P_l
    Q_l = struct[0].Q_l
    P_cig = struct[0].P_cig
    Q_cig = struct[0].Q_cig
    v_ref = struct[0].v_ref
    
    # Dynamical states:
    delta = struct[0].x[0,0]
    omega = struct[0].x[1,0]
    e1q = struct[0].x[2,0]
    e1d = struct[0].x[3,0]
    v_c = struct[0].x[4,0]
    p_m = struct[0].x[5,0]
    xi_m = struct[0].x[6,0]
    
    # Algebraic states:
    i_d = struct[0].y_run[0,0]
    i_q = struct[0].y_run[1,0]
    v_1 = struct[0].y_run[2,0]
    theta_1 = struct[0].y_run[3,0]
    P_t = struct[0].y_run[4,0]
    Q_t = struct[0].y_run[5,0]
    v_f = struct[0].y_run[6,0]
    p_m_ref = struct[0].y_run[7,0]
    v_0 = struct[0].y_run[8,0]
    theta_0 = struct[0].y_run[9,0]
    
    # Differential equations:
    if mode == 2:


        struct[0].f[0,0] = -K_delta*delta + Omega_b*(omega - omega_s)
        struct[0].f[1,0] = (-D*(omega - omega_s) - i_d*(R_a*i_d + v_1*sin(delta - theta_1)) - i_q*(R_a*i_q + v_1*cos(delta - theta_1)) + p_m)/(2*H)
        struct[0].f[2,0] = (-e1q - i_d*(-X1d + X_d) + v_f)/T1d0
        struct[0].f[3,0] = (-e1d + i_q*(-X1q + X_q))/T1q0
        struct[0].f[4,0] = (v_1 - v_c)/T_r
        struct[0].f[5,0] = (-p_m + p_m_ref)/T_m
        struct[0].f[6,0] = omega - 1
    
    # Algebraic equations:
    if mode == 3:


        struct[0].g[0,0] = R_a*i_q + X1d*i_d - e1q + v_1*cos(delta - theta_1)
        struct[0].g[1,0] = R_a*i_d - X1q*i_q - e1d + v_1*sin(delta - theta_1)
        struct[0].g[2,0] = P_cig + P_t*S_n + v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g[3,0] = Q_cig + Q_t*S_n + v_0*v_1*cos(theta_0 - theta_1)/X_l - v_1**2/X_l
        struct[0].g[4,0] = -P_t + i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].g[5,0] = -Q_t + i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].g[6,0] = K_a*(-v_c + v_pss + v_ref) - v_f
        struct[0].g[7,0] = -K_sec*xi_m - p_m_ref - (omega - 1)/Droop
        struct[0].g[8,0] = -P_l - v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g[9,0] = -Q_l - v_0**2/X_l + v_0*v_1*cos(theta_0 - theta_1)/X_l
    
    # Outputs:
    if mode == 3:

        struct[0].h[0,0] = p_m
        struct[0].h[1,0] = P_l
        struct[0].h[2,0] = Q_l
        struct[0].h[3,0] = P_cig
        struct[0].h[4,0] = Q_cig
    

    if mode == 10:

        struct[0].Fx[0,0] = -K_delta
        struct[0].Fx[0,1] = Omega_b
        struct[0].Fx[1,0] = (-i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fx[1,1] = -D/(2*H)
        struct[0].Fx[1,5] = 1/(2*H)
        struct[0].Fx[2,2] = -1/T1d0
        struct[0].Fx[3,3] = -1/T1q0
        struct[0].Fx[4,4] = -1/T_r
        struct[0].Fx[5,5] = -1/T_m
        struct[0].Fx[6,1] = 1

    if mode == 11:

        struct[0].Fy[1,0] = (-2*R_a*i_d - v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fy[1,1] = (-2*R_a*i_q - v_1*cos(delta - theta_1))/(2*H)
        struct[0].Fy[1,2] = (-i_d*sin(delta - theta_1) - i_q*cos(delta - theta_1))/(2*H)
        struct[0].Fy[1,3] = (i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fy[2,0] = (X1d - X_d)/T1d0
        struct[0].Fy[2,6] = 1/T1d0
        struct[0].Fy[3,1] = (-X1q + X_q)/T1q0
        struct[0].Fy[4,2] = 1/T_r
        struct[0].Fy[5,7] = 1/T_m

        struct[0].Gy[0,0] = X1d
        struct[0].Gy[0,1] = R_a
        struct[0].Gy[0,2] = cos(delta - theta_1)
        struct[0].Gy[0,3] = v_1*sin(delta - theta_1)
        struct[0].Gy[1,0] = R_a
        struct[0].Gy[1,1] = -X1q
        struct[0].Gy[1,2] = sin(delta - theta_1)
        struct[0].Gy[1,3] = -v_1*cos(delta - theta_1)
        struct[0].Gy[2,2] = v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[2,3] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[2,4] = S_n
        struct[0].Gy[2,8] = v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[2,9] = v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[3,2] = v_0*cos(theta_0 - theta_1)/X_l - 2*v_1/X_l
        struct[0].Gy[3,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[3,5] = S_n
        struct[0].Gy[3,8] = v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[3,9] = -v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[4,0] = v_1*sin(delta - theta_1)
        struct[0].Gy[4,1] = v_1*cos(delta - theta_1)
        struct[0].Gy[4,2] = i_d*sin(delta - theta_1) + i_q*cos(delta - theta_1)
        struct[0].Gy[4,3] = -i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1)
        struct[0].Gy[4,4] = -1
        struct[0].Gy[5,0] = v_1*cos(delta - theta_1)
        struct[0].Gy[5,1] = -v_1*sin(delta - theta_1)
        struct[0].Gy[5,2] = i_d*cos(delta - theta_1) - i_q*sin(delta - theta_1)
        struct[0].Gy[5,3] = i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].Gy[5,5] = -1
        struct[0].Gy[6,6] = -1
        struct[0].Gy[7,7] = -1
        struct[0].Gy[8,2] = -v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[8,3] = v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[8,8] = -v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[8,9] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[9,2] = v_0*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[9,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[9,8] = -2*v_0/X_l + v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[9,9] = -v_0*v_1*sin(theta_0 - theta_1)/X_l






@numba.njit(cache=True)
def Piecewise(arg):
    out = arg[0][1]
    N = len(arg)
    for it in range(N-1,-1,-1):
        if arg[it][1]: out = arg[it][0]
    return out

@numba.njit(cache=True)
def ITE(arg):
    out = arg[0][1]
    N = len(arg)
    for it in range(N-1,-1,-1):
        if arg[it][1]: out = arg[it][0]
    return out


@numba.njit(cache=True)
def Abs(x):
    return np.abs(x)



@numba.njit(cache=True) 
def daesolver(struct): 
    sin = np.sin
    cos = np.cos
    sqrt = np.sqrt
    i = 0 
    
    Dt = struct[i].Dt 

    N_x = struct[i].N_x
    N_y = struct[i].N_y
    N_z = struct[i].N_z

    decimation = struct[i].decimation 
    eye = np.eye(N_x)
    t = struct[i].t 
    t_end = struct[i].t_end 
    if struct[i].it == 0:
        run(t,struct, 1) 
        struct[i].it_store = 0  
        struct[i]['T'][0] = t 
        struct[i].X[0,:] = struct[i].x[:,0]  
        struct[i].Y[0,:] = struct[i].y_run[:,0]  
        struct[i].Z[0,:] = struct[i].h[:,0]  

    solver = struct[i].solvern 
    while t<t_end: 
        struct[i].it += 1
        struct[i].t += Dt
        
        t = struct[i].t


            
        if solver == 5: # Teapezoidal DAE as in Milano's book

            run(t,struct, 2) 
            run(t,struct, 3) 

            x = np.copy(struct[i].x[:]) 
            y = np.copy(struct[i].y_run[:]) 
            f = np.copy(struct[i].f[:]) 
            g = np.copy(struct[i].g[:]) 
            
            for iter in range(struct[i].imax):
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct,10) 
                run(t,struct,11) 
                
                x_i = struct[i].x[:] 
                y_i = struct[i].y_run[:]  
                f_i = struct[i].f[:] 
                g_i = struct[i].g[:]                 
                F_x_i = struct[i].Fx[:,:]
                F_y_i = struct[i].Fy[:,:] 
                G_x_i = struct[i].Gx[:,:] 
                G_y_i = struct[i].Gy[:,:]                

                A_c_i = np.vstack((np.hstack((eye-0.5*Dt*F_x_i, -0.5*Dt*F_y_i)),
                                   np.hstack((G_x_i,         G_y_i))))
                     
                f_n_i = x_i - x - 0.5*Dt*(f_i+f) 
                # print(t,iter,g_i)
                Dxy_i = np.linalg.solve(-A_c_i,np.vstack((f_n_i,g_i))) 
                
                x_i = x_i + Dxy_i[0:N_x]
                y_i = y_i + Dxy_i[N_x:(N_x+N_y)]

                struct[i].x[:] = x_i
                struct[i].y_run[:] = y_i

        # [f_i,g_i,F_x_i,F_y_i,G_x_i,G_y_i] =  smib_transient(x_i,y_i,u);
        
        # A_c_i = [[eye(N_x)-0.5*Dt*F_x_i, -0.5*Dt*F_y_i],
        #          [                G_x_i,         G_y_i]];
             
        # f_n_i = x_i - x - 0.5*Dt*(f_i+f);
        
        # Dxy_i = -A_c_i\[f_n_i.',g_i.'].';
        
        # x_i = x_i + Dxy_i(1:N_x);
        # y_i = y_i + Dxy_i(N_x+1:N_x+N_y);
                
                xy = np.vstack((x_i,y_i))
                max_relative = 0.0
                for it_var in range(N_x+N_y):
                    abs_value = np.abs(xy[it_var,0])
                    if abs_value < 0.001:
                        abs_value = 0.001
                                             
                    relative_error = np.abs(Dxy_i[it_var,0])/abs_value
                    
                    if relative_error > max_relative: max_relative = relative_error
                    
                if max_relative<struct[i].itol:
                    
                    break
                
                # if iter>struct[i].imax-2:
                    
                #     print('Convergence problem')

            struct[i].x[:] = x_i
            struct[i].y_run[:] = y_i
                
        # channels 
        it_store = struct[i].it_store
        if struct[i].it >= it_store*decimation: 
            struct[i]['T'][it_store+1] = t 
            struct[i].X[it_store+1,:] = struct[i].x[:,0] 
            struct[i].Y[it_store+1,:] = struct[i].y_run[:,0]
            struct[i].Z[it_store+1,:] = struct[i].h[:,0]
            struct[i].iters[it_store+1,0] = iter
            struct[i].it_store += 1 
            
    struct[i].t = t

    return t





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep  5 20:43:46 2018

@author: jmmauricio
"""

import numpy as np
import pandas as pd
import scipy
import matplotlib.pyplot as plt

from matplotlib.patches import Circle, Wedge, Polygon, Rectangle

def eval_A(system):
    
    Fx = system.struct[0].Fx
    Fy = system.struct[0].Fy
    Gx = system.struct[0].Gx
    Gy = system.struct[0].Gy
    
    A = Fx - Fy @ np.linalg.solve(Gy,Gx)
    
    system.A = A
    
    return A

def A_eval(system):
    system.jac_run_eval()
    N_x = system.N_x
    Fx = system.jac_run[:N_x,:N_x]
    Fy = system.jac_run[:N_x,N_x:]
    Gx = system.jac_run[N_x:,:N_x]
    Gy = system.jac_run[N_x:,N_x:]
    
    A = Fx - Fy @ np.linalg.solve(Gy,Gx)
    
    system.A = A
    
    return A


def eval_ss(system):
    '''
    
    Parameters
    ----------
    system : system class
        object.

    Returns
    -------
    A : np.array
        System A matrix.

    # DAE system        
    dx = f(x,y,u)
     0 = g(x,y,u)
     z = h(x,y,u)
     
    # system linealization
    Δdx = Fx*Δx + Fy*Δy + Fu*Δu
      0 = Gx*Δx + Gy*Δy + Gu*Δu
     Δz = Hx*Δx + Hy*Δy + Hu*Δu
    
    Δy = -inv(Gy)*Gx*Dx - inv(Gy)*Gu*Du
                                     
    Δdx = Fx*Dx - Fy*inv(Gy*Gx)*Δx - Fy*inv(Gy)*Gu*Δu + Fu*Δu           
    Δdx = (Fx - Fy*inv(Gy*Gx))*Δx + (Fu - Fy*inv(Gy)*Gu)*Δu
    

    Δz = Hx*Dx + Hy*Δy + Hu*Δu
    Δz = Hx*Dx - Hy*inv(Gy)*(Gx*Δx) - Hy*inv(Gy)*Gu*Du + Hu*Δu
    Δz = (Hx - Hy*inv(Gy)*(Gx))*Δx + (Hu - Hy*inv(Gy)*Gu)*Δu


    '''
    Fx = system.struct[0].Fx
    Fy = system.struct[0].Fy
    Gx = system.struct[0].Gx
    Gy = system.struct[0].Gy
    
    Fu = system.struct[0].Fu
    Gu = system.struct[0].Gu    
    
    Hx = system.struct[0].Hx
    Hy = system.struct[0].Hy  
    Hu = system.struct[0].Hu 
    
    A = Fx - Fy @ np.linalg.solve(Gy,Gx)
    B = Fu - Fy @ np.linalg.solve(Gy,Gu)
    C = Hx - Hy @ np.linalg.solve(Gy,Gx)
    D = Hu - Hy @ np.linalg.solve(Gy,Gu)
    
    system.A = A
    system.B = B
    system.C = C
    system.D = D
    
    return A



def eval_A_ini(system):
    
    Fx = system.struct[0].Fx_ini
    Fy = system.struct[0].Fy_ini
    Gx = system.struct[0].Gx_ini
    Gy = system.struct[0].Gy_ini
    
    A = Fx - Fy @ np.linalg.solve(Gy,Gx)
     
    system.A = A
    
    return A


def damp_report(system):
    eig,eigv = np.linalg.eig(system.A)
    omegas = eig.imag
    sigmas = eig.real

    freqs = np.abs(omegas/(2*np.pi))
    zetas = -sigmas/np.sqrt(sigmas**2+omegas**2)
    
    string = ''
    string += f' Mode'.ljust(10, ' ') 
    string += f' Real'.ljust(10, ' ') 
    string += f' Imag'.ljust(10, ' ') 
    string += f' Freq.'.ljust(10, ' ') 
    string += f' Damp'.ljust(10, ' ') 
    string += '\n'

    N_x = len(eig)
    for it in range(N_x):
        r = eig[it].real
        i = eig[it].imag
        string += f'{it+1:0d}  '.rjust(10, ' ') 
        string += f'{r:0.4f}  '.rjust(10, ' ') 
        string += f'{i:0.4f}j'.rjust(10, ' ') 
        string += f'{freqs[it]:0.4f}'.rjust(10, ' ') 
        string += f'{zetas[it]:0.4f}'.rjust(10, ' ') 

        string += '\n'
        #'\t{i:0.4f}\t{freqs[it]:0.3f}\t{zetas[it]:0.4f}\n'
    
    columns = ['Real','Imag','Freq.','Damp']     
    modes = [f'Mode {it+1}' for it in range(N_x)]
    eig_df = pd.DataFrame(data={'Real':eig.real,'Imag':eig.imag,  'Freq.':freqs,     'Damp':zetas},index=modes)
    
    system.eigvalues = eig
    system.eigvectors = eigv
    return eig_df
 

def participation(system, method='kundur'):
    '''
    As in PSAT but with the notation from Kundur
    
    ''' 
    
    eig,V = np.linalg.eig(system.A)
    W = np.linalg.inv(V)
    
    N_x = len(eig)
    
    participation_matrix = np.zeros((N_x,N_x))
    
    if method=='milano':
        for i in range(N_x):
            for j in range(N_x):
    
                den = np.sum(np.abs(W[j,:])*np.abs(V[:,j]))
                participation_matrix[i,j] =  np.abs(W[i,j])*np.abs(V[j,i])/den
                
        
        modes = [f'Mode {it+1}' for it in range(N_x)]
        participation_df = pd.DataFrame(data=participation_matrix,index=modes, columns=system.x_list)

    if method=='milano_transpose':
        for i in range(N_x):
            for j in range(N_x):
    
                den = np.sum(np.abs(W[j,:])*np.abs(V[:,j]))
                participation_matrix[j,i] =  np.abs(W[i,j])*np.abs(V[j,i])/den
                
        
        modes = [f'Mode {it+1}' for it in range(N_x)]
        participation_df = pd.DataFrame(data=participation_matrix,index=system.x_list, columns=modes)

    if method=='kundur':
        Phi = V
        Psi = W
        
        participation_matrix =  Phi * Psi.T
                
        
        modes = [f'Mode {it+1}' for it in range(N_x)]
        participation_df = pd.DataFrame(data=participation_matrix,index=system.x_list, columns=modes)

    
    return participation_df


def shape2df(system):
    
    eig,V = np.linalg.eig(system.A)
    W = np.linalg.inv(V)
    N_x = W.shape[0]

    participation_matrix = np.zeros((N_x,N_x))

    for i in range(N_x):
        for j in range(N_x):
            den = np.sum(np.abs(W[j,:])*np.abs(V[:,j]))
            participation_matrix[i,j] =  np.abs(W[i,j])*np.abs(V[j,i])/den
            
    modules = np.abs(participation_matrix)
    degs = np.rad2deg(np.angle(W))
    W_row = []
    W_list =[]
    N_x = W.shape[0]
    for irow in range(N_x):
        W_row = []
        for icol in range(N_x):
            W_row += [f'{modules[irow,icol]:0.2f}∠{degs[irow,icol]:5.1f}']
        W_list += [W_row]

    modes = [f'Mode {it+1}' for it in range(N_x)]

    df_shapes_report = pd.DataFrame(data=W_list,index=modes, columns=system.x_list)
    df_shapes = pd.DataFrame(data=modules*np.exp(1j*np.angle(W)),index=modes, columns=system.x_list)
    
    system.participation_matrix = participation_matrix
    system.shape_modules = modules
    system.shape_degs = degs
    system.modes_id = modes
    system.df_shapes = df_shapes
    
    return df_shapes_report
    # for it in range(N_x):
    #     r = eig[it].real
    #     i = eig[it].imag
    #     string += f'{r:0.4f}  '.rjust(10, ' ') 
    #     string += f'{i:0.4f}j'.rjust(10, ' ') 
    #     string += f'{freqs[it]:0.4f}'.rjust(10, ' ') 
    #     string += f'{zetas[it]:0.4f}'.rjust(10, ' ') 

    #     string += '\n'
    #     #'\t{i:0.4f}\t{freqs[it]:0.3f}\t{zetas[it]:0.4f}\n'
        
    # return string
    

def lqr(A,B,Q,R):
    """Solve the continuous time lqr controller.
     
    dx/dt = A x + B u
     
    cost = integral x.T*Q*x + u.T*R*u
    
    https://www.mwm.im/lqr-controllers-with-python/
    """
    #ref Bertsekas, p.151
     
    #first, try to solve the ricatti equation
    X = np.matrix(scipy.linalg.solve_continuous_are(A, B, Q, R))
     
    #compute the LQR gain
    K = np.matrix(scipy.linalg.inv(R)*(B.T*X))
     
    eigVals, eigVecs = scipy.linalg.eig(A-B*K)
     
    return K, X, eigVals



def plot_eig(grid, x_min='',x_max='',y_min='',y_max=''):
    fig,axes = plt.subplots()

    
    pi2 = 2*np.pi

    #damping zones
    ## damping zone > 10%
    e_real = -10000.0
    damp = 0.1
    e_imag = np.sqrt((e_real/damp)**2-e_real**2)/pi2
    poly = Polygon([(0,0),(e_real,0),(e_real,e_imag)], facecolor='#e1f2e1',zorder=0)
    axes.add_patch(poly)
    #axes.text(-0.85, 0.5, '$\zeta>10\%$', fontsize=12)

    ## damping zone from  5% to 10%
    axes.plot([0,e_real],[0, e_imag],'--', color='k',zorder=0)
    axes.plot([0,e_real],[0,-e_imag],'--', color='k',zorder=0)
    poly = Polygon([(0,0),(e_real,e_imag),(0,e_imag)], facecolor='#fae8ce',zorder=0)
    axes.add_patch(poly)
    #axes.text(-0.87, 9.0/pi2, '$10\% <\zeta<5\%$', fontsize=12)

    ## damping zone <5%
    e_real = -100000.0
    damp = 0.05
    e_imag = np.sqrt((e_real/damp)**2-e_real**2)/pi2
    axes.plot([0,e_real],[0, e_imag],'--', color='k',zorder=0)
    axes.plot([0,e_real],[0,-e_imag],'--', color='k',zorder=0)
    poly = Polygon([(0,0),(e_real,e_imag),(0,e_imag)], facecolor='#f7dcda',zorder=0)
    axes.add_patch(poly)
    #axes.text(-0.37, 9.0/pi2, '$\zeta<$5 %', fontsize=12)



    axes.plot([0,0],[-20,20],'-', color='k', lw=4)

    if x_min == '': x_min = np.min(grid.eigvalues.real)*1.1
    if x_max == '': x_max = np.max(grid.eigvalues.real)
    if y_min == '': y_min = np.min(grid.eigvalues.imag)/(2*np.pi)*1.1
    if y_max == '': y_max = np.max(grid.eigvalues.imag)/(2*np.pi)*1.1
        
    axes.set_xlim((x_min,x_max))
    axes.set_ylim((y_min,y_max))
    axes.grid(True)
    axes.set_xlabel('Real')
    axes.set_ylabel('Imag$/2\pi$ (Hz)')

    axes.plot(grid.eigvalues.real,grid.eigvalues.imag/(2*np.pi),'o')
    fig.tight_layout()
    
    return fig
    
def add_arrow(string,name='arrow',scale=1,angle=0,center_x=0,center_y=0,color="#337ab7"):
    trans_x = center_x*(1-scale)
    trans_y = center_y*(1-scale)
    string += f'<path d="M {center_x},{center_y} h 1.32292 v -52.916665 h 1.32291 l -2.64583,-5.291666 -2.64583,5.291666 h 1.32291 v 52.916665 z" ' 
    string += f'fill="{color}" id="arrow_1" class="tooltip-trigger" data-tooltip-text="{name}"  fill-opacity="0.5" '
    string += f'stroke-width="0" transform="translate({trans_x},{trans_y}) scale({scale})  rotate({angle},{center_x},{center_y})" '   
    string += f'/>' 
    return string


def plot_shapes(grid,mode='Mode 13',states=['omega_1','omega_1','omega_3','omega_4'], plot_scale=3):
    svg_arrows = ''
    shapes = grid.df_shapes.loc[mode][states]

    max_module = shapes.abs().max()
    for shape,name in zip(shapes,shapes.index):
        module = np.abs(shape)
        angle = np.angle(shape,deg=True)
        svg_arrows = add_arrow(svg_arrows,name=name,scale=plot_scale*module/max_module,angle=angle,center_x=200,center_y=200)
        
    # from: https://www.petercollingridge.co.uk/tutorials/svg/interactive/tooltip/    
    svg_string = ''' 
        <svg xmlns="http://www.w3.org/2000/svg" 
        height="400"  width="400"
        id="tooltip-svg-5">
            <style>
                #tooltip {
                    dominant-baseline: hanging; 
                }
            </style>
            
            <circle cx="200" cy="200" r="190" stroke="#888888" stroke-width="3" fill="#DDDDDD" />

        {arrows}    
        
        <g id="tooltip" visibility="hidden" >

                <rect width="80" height="24" fill="white" rx="2" ry="2"/>
                <text x="3" y="6">Tooltip</text>
            </g>


            <script type="text/ecmascript"><![CDATA[
                (function() {
                    var svg = document.getElementById('tooltip-svg-5');
                    var tooltip = svg.getElementById('tooltip');
                    var tooltipText = tooltip.getElementsByTagName('text')[0].firstChild;
                    var triggers = svg.getElementsByClassName('tooltip-trigger');

                    for (var i = 0; i < triggers.length; i++) {
                        triggers[i].addEventListener('mousemove', showTooltip);
                        triggers[i].addEventListener('mouseout', hideTooltip);
                    }

                    function showTooltip(evt) {
                        var CTM = svg.getScreenCTM();
                        var x = (evt.clientX - CTM.e + 6) / CTM.a;
                        var y = (evt.clientY - CTM.f + 20) / CTM.d;
                        tooltip.setAttributeNS(null, "transform", "translate(" + x + " " + y + ")");
                        tooltip.setAttributeNS(null, "visibility", "visible");
                        tooltipText.data = evt.target.getAttributeNS(null, "data-tooltip-text");
                    }

                    function hideTooltip(evt) {
                        tooltip.setAttributeNS(null, "visibility", "hidden");
                    }
                })()
            ]]></script>
        </svg>
        '''
    
    svg_string = svg_string.replace('{arrows}',svg_arrows)
    
    return svg_string
    

def plot_vectors(vectors,names,colors, plot_scale=3):
    svg_arrows = ''

    for vector,name,color in zip(vectors,names,colors):
        module = np.abs(vector)
        angle = np.angle(vector,deg=True)
        svg_arrows = add_arrow(svg_arrows,name=name,scale=plot_scale*module,angle=angle,center_x=200,center_y=200,color=color)
        
    # from: https://www.petercollingridge.co.uk/tutorials/svg/interactive/tooltip/    
    svg_string = ''' 
        <svg xmlns="http://www.w3.org/2000/svg" 
        height="400"  width="400"
        id="tooltip-svg-5">
            <style>
                #tooltip {
                    dominant-baseline: hanging; 
                }
            </style>
            
            <circle cx="200" cy="200" r="190" stroke="#888888" stroke-width="3" fill="#DDDDDD" />

        {arrows}    
        
        <g id="tooltip" visibility="hidden" >

                <rect width="80" height="24" fill="white" rx="2" ry="2"/>
                <text x="3" y="6">Tooltip</text>
            </g>


            <script type="text/ecmascript"><![CDATA[
                (function() {
                    var svg = document.getElementById('tooltip-svg-5');
                    var tooltip = svg.getElementById('tooltip');
                    var tooltipText = tooltip.getElementsByTagName('text')[0].firstChild;
                    var triggers = svg.getElementsByClassName('tooltip-trigger');

                    for (var i = 0; i < triggers.length; i++) {
                        triggers[i].addEventListener('mousemove', showTooltip);
                        triggers[i].addEventListener('mouseout', hideTooltip);
                    }

                    function showTooltip(evt) {
                        var CTM = svg.getScreenCTM();
                        var x = (evt.clientX - CTM.e + 6) / CTM.a;
                        var y = (evt.clientY - CTM.f + 20) / CTM.d;
                        tooltip.setAttributeNS(null, "transform", "translate(" + x + " " + y + ")");
                        tooltip.setAttributeNS(null, "visibility", "visible");
                        tooltipText.data = evt.target.getAttributeNS(null, "data-tooltip-text");
                    }

                    function hideTooltip(evt) {
                        tooltip.setAttributeNS(null, "visibility", "hidden");
                    }
                })()
            ]]></script>
        </svg>
        '''
    
    svg_string = svg_string.replace('{arrows}',svg_arrows)
    
    return svg_string

    
def dlqr(A,B,Q,R):
    """Solve the discrete time lqr controller.
     
    x[k+1] = A x[k] + B u[k]
     
    cost = sum x[k].T*Q*x[k] + u[k].T*R*u[k]
    
    https://www.mwm.im/lqr-controllers-with-python/
    """
    #ref Bertsekas, p.151
     
    #first, try to solve the ricatti equation
    X = np.matrix(scipy.linalg.solve_discrete_are(A, B, Q, R))
     
    #compute the LQR gain
    K = np.matrix(scipy.linalg.inv(B.T*X*B+R)*(B.T*X*A))
     
    eigVals, eigVecs = scipy.linalg.eig(A-B*K)
     
    return K, X, eigVals


def discretise_time(A, B, dt):
    '''Compute the exact discretization of the continuous system A,B.
    
    Goes from a description 
     d/dt x(t) = A*x(t) + B*u(t)
     u(t)  = ud[k] for t in [k*dt, (k+1)*dt)
    to the description
     xd[k+1] = Ad*xd[k] + Bd*ud[k]
    where
     xd[k] := x(k*dt)
     
    Returns: Ad, Bd
    
    from: https://github.com/markwmuller/controlpy/blob/master/controlpy/analysis.py
    '''
    
    nstates = A.shape[0]
    ninputs = B.shape[1]

    M = np.matrix(np.zeros([nstates+ninputs,nstates+ninputs]))
    M[:nstates,:nstates] = A
    M[:nstates, nstates:] = B
    
    Md = scipy.linalg.expm(M*dt)
    Ad = Md[:nstates, :nstates]
    Bd = Md[:nstates, nstates:]

    return Ad, Bd


def acker(A,B,poles):
    '''
    This function is a copy from the original in: https://github.com/python-control/python-control
    but it allows to work with complex A and B matrices. It is experimental and the original should be
    considered
    
    
    ----------
    A : numpy array_like (complex can be used)
        Dynamics amatrix of the system
    B : numpy array_like (complex can be used)
        Input matrix of the system
    poles : numpy array_like
        Desired eigenvalue locations.

    Returns
    -------
    K : numpy array_like
        Gain such that A - B K has eigenvalues given in p.

    '''
    
    
    N_x = np.shape(A)[0]
    
    ctrb = np.hstack([B] + [np.dot(np.linalg.matrix_power(A, i), B)
                                             for i in range(1, N_x)])

    # Compute the desired characteristic polynomial
    p = np.real(np.poly(poles))

    n = np.size(p)
    pmat = p[n-1] * np.linalg.matrix_power(A, 0)
    for i in np.arange(1,n):
        pmat = pmat + np.dot(p[n-i-1], np.linalg.matrix_power(A, i))
    K = np.linalg.solve(ctrb, pmat)

    K = K[-1][:]                # Extract the last row            # Extract the last row
    
    return K    

def right2df(system):
    
    eig,V = np.linalg.eig(system.A)
    W = np.linalg.inv(V)
    N_x = W.shape[0]

    modules = np.abs(V)
    degs = np.rad2deg(np.angle(V))
    W_row = []
    W_list =[]
    N_x = W.shape[0]
    for irow in range(N_x):
        W_row = []
        for icol in range(N_x):
            W_row += [f'{modules[icol,irow]:0.2f}∠{degs[icol,irow]:5.1f}']
        W_list += [W_row]

    modes = [f'Mode {it+1}' for it in range(N_x)]

    df_report = pd.DataFrame(data=W_list,index=modes, columns=system.x_list)
    df = pd.DataFrame(data=modules*np.exp(1j*np.angle(W)),index=modes, columns=system.x_list)
    
    system.shape_modules = modules
    system.shape_degs = degs
    system.modes_id = modes
    system.df = df
    
    return df_report

def left2df(system):
    
    eig,V = np.linalg.eig(system.A)
    W = np.linalg.inv(V)
    N_x = W.shape[0]

    modules = np.abs(W)
    degs = np.rad2deg(np.angle(W))
    W_row = []
    W_list =[]
    N_x = W.shape[0]
    for irow in range(N_x):
        W_row = []
        for icol in range(N_x):
            W_row += [f'{modules[irow,icol]:0.2f}∠{degs[irow,icol]:5.1f}']
        W_list += [W_row]

    modes = [f'Mode {it+1}' for it in range(N_x)]

    df_report = pd.DataFrame(data=W_list,index=modes, columns=system.x_list)
    df = pd.DataFrame(data=modules*np.exp(1j*np.angle(W)),index=modes, columns=system.x_list)
    
    system.shape_modules = modules
    system.shape_degs = degs
    system.modes_id = modes
    system.df = df
    
    return df_report


def pi_design(a,b,zeta,omega):
    '''
    G = 1/(a*s + b)    
    PI = (K_p + K_i/s) = (K_p*s + K_i)/s -> tf([K_p,K_i],[1,0])
    K_i = K_p/T_p
    T_p = K_p/K_i
    
    Example:
    
    L = 4e-3
    R = 0.1

    K_p,K_i = pi_design(L,R,1.0,2*np.pi*10)

    G_planta = ctrl.tf([1], [L,R])
    G_pi = ctrl.tf([K_p,K_i],[1,0])

    H = G_planta*G_pi/(1 + G_planta*G_pi)
    ctrl.bode_plot(H);
    
    '''
    
    K_p = 2*zeta*omega*a-b
    K_i = a*omega**2
    
    return K_p,K_i

def lead_design(angle,omega, verbose=False):
    '''
    Desing lead lag compensator: G = (T_1*s + 1)/(T_2*s + 1)
    
    alpha = (1+np.sin(angle))/(1-np.sin(angle))
    
    T_2 = 1/(omega*np.sqrt(alpha))
    T_1 = alpha*T_2
    

    G_ll = ctrl.tf([T_1,1],[T_2,1])

    ctrl.bode_plot(G_ll);
    
    '''
    
    alpha = (1+np.sin(angle))/(1-np.sin(angle))
    T_2 = 1/(omega*np.sqrt(alpha))
    T_1 = alpha*T_2
    
    if verbose:
        G_omega = (T_1*1j*omega+1)/(T_2*1j*omega+1)
        print(f'Gain at omega: {np.abs(G_omega)}')
    
    return T_1,T_2
    
     
if __name__ == "__main__":
    
    class sys_class():
        
        def __init__(self): 
     
            
            self.A = np.array([[ 0.00000000e+00,  3.14159265e+02,  0.00000000e+00,
                 0.00000000e+00,  0.00000000e+00],
               [-4.43027144e-01, -1.42857143e-01, -1.44867645e-01,
                 1.69999669e-01,  0.00000000e+00],
               [-1.80455968e-01,  0.00000000e+00, -6.64265905e-01,
                -2.31113959e-03, -2.50000000e+01],
               [ 1.32845727e+00,  0.00000000e+00,  1.35913375e-02,
                -2.58565604e+00,  0.00000000e+00],
               [-3.90435895e-01,  0.00000000e+00,  2.69427382e+00,
                 4.81098835e-01, -2.00000000e+01]])


            self.A = np.array([[      0,  314.1593,        0,         0,         0,         0],
                               [-0.1594,         0,        0,         0,   -0.1730,    0.0600],
                               [-0.2258,         0,  -0.0857,         0,   -0.2669,   -0.0011],
                               [ 0.4611,         0,        0,   -1.0000,    0.0063,   -1.4894],
                               [-2.9728,         0,  33.3333,         0,  -36.8474,   -0.0145],
                               [ 2.6216,         0,        0,   14.2857,    0.0360,  -22.7542]])
            
            self.A = np.array([[0,314.16,0,0,0,0],
                                [-0.15935,0,0,0,-0.17302,0.060008],
                                [-0.22579,0,-0.085744,0,-0.26689,-0.0011041],
                                [0.46108,0,0,-1,0.0063364,-1.4894],
                                [-2.9728,0,33.333,0,-36.847,-0.014537],
                                [2.6216,0,0,14.286,0.036027,-22.754]])
            
            self.x_list = ['delta_Syn_1','omega_Syn_1','e1q_Syn_1','e1d_Syn_1','e2q_Syn_1','e2d_Syn_1']

            
            # Example 12.4 (AVR+PSS):
            self.A = np.array([[     0, -0.1092, -0.1236, 0, 0, 0],
                               [376.99,       0,       0, 0, 0, 0],
                               [     0, -0.1938, -0.4229, -27.3172, 0, 27.3172],
                               [      0,-7.3125,   20.8391, -50.0, 0,0],
                               [     0,-1.0372, -1.1738,0, -0.7143,0],
                               [     0,-4.8404, -5.4777,0,26.9697, -30.3030]])
           
            # Example 12.4 (AVR+PSS):
            self.x_list = ['Domega_r','Ddelta','Dpsi_fd','Dv_1','Dv_2','Dv_s']
            

    sys = sys_class()
    print(damp_report(sys))
    PF = participation(sys)
    df_shape = shape2df(sys)

    
'''    
STATE MATRIX EIGENVALUES

Eigevalue      Most Associated States      Real part      Imag. Part     Pseudo-Freq.   Frequency      

Eig As #1      e2q_Syn_1                  -36.4944        0              0              0             
Eig As #2      e2d_Syn_1                  -21.6378        0              0              0             
Eig As #3      omega_Syn_1, delta_Syn_1   -0.25606        6.633          1.0557         1.0565        
Eig As #4      omega_Syn_1, delta_Syn_1   -0.25606       -6.633          1.0557         1.0565        
Eig As #5      e1d_Syn_1                  -1.9568         0              0              0             
Eig As #6      e1q_Syn_1                  -0.08604        0              0              0             

PARTICIPATION FACTORS (Euclidean norm)

               delta_Syn_1    omega_Syn_1    e1q_Syn_1      e1d_Syn_1      e2q_Syn_1      

Eig As #1      0.00283        0.00283        0.00641        0              0.98792       
Eig As #2      0.00344        0.00344        1e-05          0.04673        0.0001        
Eig As #3      0.4805         0.4805         0.01663        0.00586        0.0091        
Eig As #4      0.4805         0.4805         0.01663        0.00586        0.0091        
Eig As #5      0.0023         0.0023         0.00474        0.94518        7e-05         
Eig As #6      0.0005         0.0005         0.99273        0.00576        1e-05         

PARTICIPATION FACTORS (Euclidean norm)

               e2d_Syn_1      

Eig As #1      2e-05         
Eig As #2      0.94629       
Eig As #3      0.00741       
Eig As #4      0.00741       
Eig As #5      0.04542       
Eig As #6      0.00051       

STATISTICS

DYNAMIC ORDER                 6             
# OF EIGS WITH Re(mu) < 0     6             
# OF EIGS WITH Re(mu) > 0     0             
# OF REAL EIGS                4             
# OF COMPLEX PAIRS            1             
# OF ZERO EIGS                0 
'''

['delta_Syn_1','omega_Syn_1','e1q_Syn_1','e1d_Syn_1','e2q_Syn_1','e2d_Syn_1']



import numpy as np
import numba



@numba.njit(cache=True) 
def interp(t,T,X):
    
    if t<=T[-1]:
        t_idx = np.argmax(T>t)    
        t_1 = T[t_idx-1] 
        t_2 = T[t_idx] 
        x_1 = X[t_idx-1] 
        x_2 = X[t_idx] 
    
        x = (x_2-x_1)/(t_2-t_1)*(t-t_1)+x_1
    else:
        x = X[-1]
    return x


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 15 23:38:58 2021

@author: jmmau
"""

import pytest
import numpy as np
import sympy as sym
import pydae.build_cffi as db
 
 
    
def test_pendulum_builder():
    params_dict = {'L':5.21,'G':9.81,'M':10.0,'K_d':1e-6}


    u_ini_dict = {'theta':np.deg2rad(5.0)}  # for the initialization problem
    u_run_dict = {'f_x':10}  # for the running problem (here initialization and running problem are the same)
    
    #u_ini_dict = {'theta':10}  # for the initialization problem
    u_run_dict = {'f_x':10}  # for the running problem (here initialization and running problem are the same)
    
    
    
    x_list = ['pos_x','pos_y','v_x','v_y']    # [inductor current, PI integrator]
    y_ini_list = ['lam','f_x'] # for the initialization problem
    y_run_list = ['lam','theta'] # for the running problem (here initialization and running problem are the same)
    
    sys_vars = {'params':params_dict,
                'u_list':u_run_dict,
                'x_list':x_list,
                'y_list':y_run_list}
    
    exec(db.sym_gen_str())  # exec to generate the required symbolic varables and constants
    
    dpos_x = v_x
    dpos_y = v_y
    dv_x = (-2*pos_x*lam + f_x - K_d*v_x)/M
    dv_y = (-M*G - 2*pos_y*lam - K_d*v_y)/M   
    
    g_1 = pos_x**2 + pos_y**2 - L**2 -lam*1e-6
    g_2 = -theta + sym.atan2(pos_x,-pos_y)

    sys = {'name':'pendulum',
           'params_dict':params_dict,
           'f_list':[dpos_x,dpos_y,dv_x,dv_y],
           'g_list':[g_1,g_2],
           'x_list':x_list,
           'y_ini_list':y_ini_list,
           'y_run_list':y_run_list,
           'u_run_dict':u_run_dict,
           'u_ini_dict':u_ini_dict,
           'h_dict':{'g_1':g_1,'PE':M*G*pos_y,'KE':0.5*M*(v_x**2+v_y**2),'theta':theta}}
    
    sys = db.build(sys)


def test_pendulum_ini():
    
    import pendulum
    
    pend = pendulum.pendulum_class()
    M = 30.0
    L = 5.21
    pend.ini({'f_x':0,'M':M,'L':L,'theta':np.deg2rad(-5)},-5)
    
    expected_results ={'pos_x' : -0.4541,
                        'pos_y' : -5.19,
                        'v_x' :     0.0,
                        'v_y' :     0.0,
                        'lam' : 28.35,
                        'g_1' : -3.553e-15,
                        'PE' : -1527.46911,
                        'KE' :     0,
                        'theta' : -0.08727,
                        'f_x'  :-25.7479}

    
    
    for item in expected_results:
        item_computed_value = pend.get_value(item)
        assert np.abs(item_computed_value-expected_results[item])<0.01
        
    
    del pendulum 



if __name__ == "__main__":
    
    #test_pendulum_builder()
    test_pendulum_ini()


import numpy as np
import numba
import scipy.optimize as sopt
import json

sin = np.sin
cos = np.cos
atan2 = np.arctan2
sqrt = np.sqrt 
sign = np.sign 
exp = np.exp


class {name}_class: 

    def __init__(self): 

        self.t_end = 10.000000 
        self.Dt = 0.0010000 
        self.decimation = 10.000000 
        self.itol = 1e-6 
        self.Dt_max = 0.001000 
        self.Dt_min = 0.001000 
        self.solvern = 5 
        self.imax = 100 
        self.N_x = {N_x}
        self.N_y = {N_y} 
        self.N_z = {N_z} 
        self.N_store = 10000 
        self.params_list = {params_list} 
        self.params_values_list  = {params_values_list} 
        self.inputs_ini_list = {inputs_ini_list} 
        self.inputs_ini_values_list  = {inputs_ini_values_list} 
        self.inputs_run_list = {inputs_run_list} 
        self.inputs_run_values_list = {inputs_run_values_list} 
        self.outputs_list = {outputs_list} 
        self.x_list = {x_list} 
        self.y_run_list = {y_run_list} 
        self.xy_list = self.x_list + self.y_run_list 
        self.y_ini_list = {y_ini_list} 
        self.xy_ini_list = self.x_list + self.y_ini_list 
        self.t = 0.0
        self.it = 0
        self.it_store = 0
        self.xy_prev = np.zeros((self.N_x+self.N_y,1))
        self.initialization_tol = 1e-6
        self.N_u = len(self.inputs_run_list) 
        self.sopt_root_method='hybr'
        self.sopt_root_jac=True
        self.u_ini_list = self.inputs_ini_list
        self.u_ini_values_list = self.inputs_ini_values_list
        self.u_run_list = self.inputs_run_list
        self.u_run_values_list = self.inputs_run_values_list
        self.N_u = len(self.u_run_list)
        Fx_ini_rows,Fx_ini_cols,Fy_ini_rows,Fy_ini_cols,Gx_ini_rows,Gx_ini_cols,Gy_ini_rows,Gy_ini_cols = nonzeros()

        self.Fx_ini_rows = np.array(Fx_ini_rows) 
        if len(Fx_ini_rows) == 1: 
            self.Fx_ini_rows = np.array([[Fx_ini_rows]]).reshape(1,) 
            self.Fx_ini_cols = np.array([[Fx_ini_cols]]).reshape(1,)  
            
        self.Fx_ini_cols = np.array(Fx_ini_cols)
        self.Fy_ini_rows = np.array(Fy_ini_rows)        
        self.Fy_ini_cols = np.array(Fy_ini_cols)
        self.Gx_ini_rows = np.array(Gx_ini_rows)        
        self.Gx_ini_cols = np.array(Gx_ini_cols)
        self.Gy_ini_rows = np.array(Gy_ini_rows)        
        self.Gy_ini_cols = np.array(Gy_ini_cols)
        
        
        self.yini2urun = list(set(self.inputs_run_list).intersection(set(self.y_ini_list)))
        self.uini2yrun = list(set(self.y_run_list).intersection(set(self.inputs_ini_list)))

        self.update() 
                
    def update(self): 

        self.N_steps = int(np.ceil(self.t_end/self.Dt)) 
        dt = [  
              ('t_end', np.float64),
              ('Dt', np.float64),
              ('decimation', np.float64),
              ('itol', np.float64),
              ('Dt_max', np.float64),
              ('Dt_min', np.float64),
              ('solvern', np.int64),
              ('imax', np.int64),
              ('N_steps', np.int64),
              ('N_store', np.int64),
              ('N_x', np.int64),
              ('N_y', np.int64),
              ('N_z', np.int64),
              ('t', np.float64),
              ('it', np.int64),
              ('it_store', np.int64),
              ('idx', np.int64),
              ('idy', np.int64),
              ('f', np.float64, (self.N_x,1)),
              ('x', np.float64, (self.N_x,1)),
              ('x_0', np.float64, (self.N_x,1)),
              ('g', np.float64, (self.N_y,1)),
              ('y_run', np.float64, (self.N_y,1)),
              ('y_ini', np.float64, (self.N_y,1)),
              ('u_run', np.float64, (self.N_u,1)),
              ('y_0', np.float64, (self.N_y,1)),
              ('h', np.float64, (self.N_z,1)),
              ('Fx', np.float64, (self.N_x,self.N_x)),
              ('Fy', np.float64, (self.N_x,self.N_y)),
              ('Gx', np.float64, (self.N_y,self.N_x)),
              ('Gy', np.float64, (self.N_y,self.N_y)),
              ('Fu', np.float64, (self.N_x,self.N_u)),
              ('Gu', np.float64, (self.N_y,self.N_u)),
              ('Hx', np.float64, (self.N_z,self.N_x)),
              ('Hy', np.float64, (self.N_z,self.N_y)),
              ('Hu', np.float64, (self.N_z,self.N_u)),
              ('Fx_ini', np.float64, (self.N_x,self.N_x)),
              ('Fy_ini', np.float64, (self.N_x,self.N_y)),
              ('Gx_ini', np.float64, (self.N_y,self.N_x)),
              ('Gy_ini', np.float64, (self.N_y,self.N_y)),
              ('T', np.float64, (self.N_store+1,1)),
              ('X', np.float64, (self.N_store+1,self.N_x)),
              ('Y', np.float64, (self.N_store+1,self.N_y)),
              ('Z', np.float64, (self.N_store+1,self.N_z)),
              ('iters', np.float64, (self.N_store+1,1)),
              ('store', np.int64),
              ('Fx_ini_rows', np.int64, self.Fx_ini_rows.shape),
              ('Fx_ini_cols', np.int64, self.Fx_ini_cols.shape),
              ('Fy_ini_rows', np.int64, self.Fy_ini_rows.shape),
              ('Fy_ini_cols', np.int64, self.Fy_ini_cols.shape),
              ('Gx_ini_rows', np.int64, self.Gx_ini_rows.shape),
              ('Gx_ini_cols', np.int64, self.Gx_ini_cols.shape),
              ('Gy_ini_rows', np.int64, self.Gy_ini_rows.shape),
              ('Gy_ini_cols', np.int64, self.Gy_ini_cols.shape),
              ('Ac_ini', np.float64, ((self.N_x+self.N_y,self.N_x+self.N_y))),   
              ('fg', np.float64, ((self.N_x+self.N_y,1))),  
             ]



        
        
        values = [
                self.t_end,                          
                self.Dt,
                self.decimation,
                self.itol,
                self.Dt_max,
                self.Dt_min,
                self.solvern,
                self.imax,
                self.N_steps,
                self.N_store,
                self.N_x,
                self.N_y,
                self.N_z,
                self.t,
                self.it,
                self.it_store,
                0,                                     # idx
                0,                                     # idy
                np.zeros((self.N_x,1)),                # f
                np.zeros((self.N_x,1)),                # x
                np.zeros((self.N_x,1)),                # x_0
                np.zeros((self.N_y,1)),                # g
                np.zeros((self.N_y,1)),                # y_run
                np.zeros((self.N_y,1)),                # y_ini
                np.zeros((self.N_u,1)),                # u_run
                np.zeros((self.N_y,1)),                # y_0
                np.zeros((self.N_z,1)),                # h
                np.zeros((self.N_x,self.N_x)),         # Fx   
                np.zeros((self.N_x,self.N_y)),         # Fy 
                np.zeros((self.N_y,self.N_x)),         # Gx 
                np.zeros((self.N_y,self.N_y)),         # Fy
                np.zeros((self.N_x,self.N_u)),         # Fu 
                np.zeros((self.N_y,self.N_u)),         # Gu 
                np.zeros((self.N_z,self.N_x)),         # Hx 
                np.zeros((self.N_z,self.N_y)),         # Hy 
                np.zeros((self.N_z,self.N_u)),         # Hu 
                np.zeros((self.N_x,self.N_x)),         # Fx_ini  
                np.zeros((self.N_x,self.N_y)),         # Fy_ini 
                np.zeros((self.N_y,self.N_x)),         # Gx_ini 
                np.zeros((self.N_y,self.N_y)),         # Fy_ini 
                np.zeros((self.N_store+1,1)),          # T
                np.zeros((self.N_store+1,self.N_x)),   # X
                np.zeros((self.N_store+1,self.N_y)),   # Y
                np.zeros((self.N_store+1,self.N_z)),   # Z
                np.zeros((self.N_store+1,1)),          # iters
                1,
                self.Fx_ini_rows,       
                self.Fx_ini_cols,
                self.Fy_ini_rows,       
                self.Fy_ini_cols,
                self.Gx_ini_rows,        
                self.Gx_ini_cols,
                self.Gy_ini_rows,       
                self.Gy_ini_cols,
                np.zeros((self.N_x+self.N_y,self.N_x+self.N_y)),  
                np.zeros((self.N_x+self.N_y,1)),
                ]  

        dt += [(item,np.float64) for item in self.params_list]
        values += [item for item in self.params_values_list]

        for item_id,item_val in zip(self.inputs_ini_list,self.inputs_ini_values_list):
            if item_id in self.inputs_run_list: continue
            dt += [(item_id,np.float64)]
            values += [item_val]

        dt += [(item,np.float64) for item in self.inputs_run_list]
        values += [item for item in self.inputs_run_values_list]

        self.struct = np.rec.array([tuple(values)], dtype=np.dtype(dt))
        
        xy0 = np.zeros((self.N_x+self.N_y,))
        self.ini_dae_jacobian_nn(xy0)
        self.run_dae_jacobian_nn(xy0)
        


    def load_params(self,data_input):

        if type(data_input) == str:
            json_file = data_input
            self.json_file = json_file
            self.json_data = open(json_file).read().replace("'",'"')
            data = json.loads(self.json_data)
        elif type(data_input) == dict:
            data = data_input

        self.data = data
        for item in self.data:
            self.struct[0][item] = self.data[item]
            if item in self.params_list:
                self.params_values_list[self.params_list.index(item)] = self.data[item]
            elif item in self.inputs_ini_list:
                self.inputs_ini_values_list[self.inputs_ini_list.index(item)] = self.data[item]
            elif item in self.inputs_run_list:
                self.inputs_run_values_list[self.inputs_run_list.index(item)] = self.data[item]
            else: 
                print(f'parameter or input {item} not found')

    def save_params(self,file_name = 'parameters.json'):
        params_dict = {}
        for item in self.params_list:
            params_dict.update({item:self.get_value(item)})

        params_dict_str = json.dumps(params_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(params_dict_str)

    def save_inputs_ini(self,file_name = 'inputs_ini.json'):
        inputs_ini_dict = {}
        for item in self.inputs_ini_list:
            inputs_ini_dict.update({item:self.get_value(item)})

        inputs_ini_dict_str = json.dumps(inputs_ini_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(inputs_ini_dict_str)

    def ini_problem(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        if self.compile:
            ini(self.struct,2)
            ini(self.struct,3)       
        else:
            ini.py_func(self.struct,2)
            ini.py_func(self.struct,3)                   
        fg = np.vstack((self.struct[0].f,self.struct[0].g))[:,0]
        return fg

    def run_problem(self,x):
        t = self.struct[0].t
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_run[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        
        if self.compile:
            run(t,self.struct,2)
            run(t,self.struct,3)
            run(t,self.struct,10)
            run(t,self.struct,11)
            run(t,self.struct,12)
            run(t,self.struct,13)
        else:
            run.py_func(t,self.struct,2)
            run.py_func(t,self.struct,3)
            run.py_func(t,self.struct,10)
            run.py_func(t,self.struct,11)
            run.py_func(t,self.struct,12)
            run.py_func(t,self.struct,13)            
        
        fg = np.vstack((self.struct[0].f,self.struct[0].g))[:,0]
        return fg
    

    def run_dae_jacobian(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_run[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        run(0.0,self.struct,10)
        run(0.0,self.struct,11)     
        run(0.0,self.struct,12)
        run(0.0,self.struct,13)
        A_c = np.block([[self.struct[0].Fx,self.struct[0].Fy],
                        [self.struct[0].Gx,self.struct[0].Gy]])
        return A_c

    def run_dae_jacobian_nn(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_run[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        run_nn(0.0,self.struct,10)
        run_nn(0.0,self.struct,11)     
        run_nn(0.0,self.struct,12)
        run_nn(0.0,self.struct,13)
 

    
    def eval_jacobians(self):

        run(0.0,self.struct,10)
        run(0.0,self.struct,11)  
        run(0.0,self.struct,12) 

        return 1


    def ini_dae_jacobian(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        if self.compile:
            ini(self.struct,10)
            ini(self.struct,11) 
        else:
            ini.py_func(self.struct,10)
            ini.py_func(self.struct,11)             
        A_c = np.block([[self.struct[0].Fx_ini,self.struct[0].Fy_ini],
                        [self.struct[0].Gx_ini,self.struct[0].Gy_ini]])
        return A_c

    def ini_dae_jacobian_nn(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        ini_nn(self.struct,10)
        ini_nn(self.struct,11)       
 

    def f_ode(self,x):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def f_odeint(self,x,t):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def f_ivp(self,t,x):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def Fx_ode(self,x):
        self.struct[0].x[:,0] = x
        run(self.struct,10)
        return self.struct[0].Fx

    def eval_A(self):
        
        Fx = self.struct[0].Fx
        Fy = self.struct[0].Fy
        Gx = self.struct[0].Gx
        Gy = self.struct[0].Gy
        
        A = Fx - Fy @ np.linalg.solve(Gy,Gx)
        
        self.A = A
        
        return A

    def eval_A_ini(self):
        
        Fx = self.struct[0].Fx_ini
        Fy = self.struct[0].Fy_ini
        Gx = self.struct[0].Gx_ini
        Gy = self.struct[0].Gy_ini
        
        A = Fx - Fy @ np.linalg.solve(Gy,Gx)
        
        
        return A
    
    def reset(self):
        for param,param_value in zip(self.params_list,self.params_values_list):
            self.struct[0][param] = param_value
        for input_name,input_value in zip(self.inputs_ini_list,self.inputs_ini_values_list):
            self.struct[0][input_name] = input_value   
        for input_name,input_value in zip(self.inputs_run_list,self.inputs_run_values_list):
            self.struct[0][input_name] = input_value  

    def simulate(self,events,xy0=0):
        
        # initialize both the ini and the run system
        self.initialize(events,xy0=xy0)
        
        # simulation run
        for event in events:  
            # make all the desired changes
            self.run([event]) 
            
        # post process
        T,X,Y,Z = self.post()
        
        return T,X,Y,Z
    

    
    def run(self,events):
        

        # simulation run
        for event in events:  
            # make all the desired changes
            for item in event:
                self.struct[0][item] = event[item]
            daesolver(self.struct)    # run until next event
            
        return 1
 
    def rtrun(self,events):
        

        # simulation run
        for event in events:  
            # make all the desired changes
            for item in event:
                self.struct[0][item] = event[item]
            self.struct[0].it_store = self.struct[0].N_store-1
            daesolver(self.struct)    # run until next event
            
            
        return 1
    
    def post(self):
        
        # post process result    
        T = self.struct[0]['T'][:self.struct[0].it_store]
        X = self.struct[0]['X'][:self.struct[0].it_store,:]
        Y = self.struct[0]['Y'][:self.struct[0].it_store,:]
        Z = self.struct[0]['Z'][:self.struct[0].it_store,:]
        iters = self.struct[0]['iters'][:self.struct[0].it_store,:]
    
        self.T = T
        self.X = X
        self.Y = Y
        self.Z = Z
        self.iters = iters
        
        return T,X,Y,Z
        
    def save_0(self,file_name = 'xy_0.json'):
        xy_0_dict = {}
        for item in self.x_list:
            xy_0_dict.update({item:self.get_value(item)})
        for item in self.y_ini_list:
            xy_0_dict.update({item:self.get_value(item)})
    
        xy_0_str = json.dumps(xy_0_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(xy_0_str)

    def load_0(self,file_name = 'xy_0.json'):
        with open(file_name) as fobj:
            xy_0_str = fobj.read()
        xy_0_dict = json.loads(xy_0_str)
    
        for item in xy_0_dict:
            if item in self.x_list:
                self.xy_prev[self.x_list.index(item)] = xy_0_dict[item]
            if item in self.y_ini_list:
                self.xy_prev[self.y_ini_list.index(item)+self.N_x] = xy_0_dict[item]
                
            
    def initialize(self,events=[{}],xy0=0,compile=True):
        '''
        

        Parameters
        ----------
        events : dictionary 
            Dictionary with at least 't_end' and all inputs and parameters 
            that need to be changed.
        xy0 : float or string, optional
            0 means all states should be zero as initial guess. 
            If not zero all the states initial guess are the given input.
            If 'prev' it uses the last known initialization result as initial guess.

        Returns
        -------
        T : TYPE
            DESCRIPTION.
        X : TYPE
            DESCRIPTION.
        Y : TYPE
            DESCRIPTION.
        Z : TYPE
            DESCRIPTION.

        '''
        
        self.compile = compile
        
        # simulation parameters
        self.struct[0].it = 0       # set time step to zero
        self.struct[0].it_store = 0 # set storage to zero
        self.struct[0].t = 0.0      # set time to zero
                    
        # initialization
        it_event = 0
        event = events[it_event]
        for item in event:
            self.struct[0][item] = event[item]
            
        
        ## compute initial conditions using x and y_ini 
        if type(xy0) == str:
            if xy0 == 'prev':
                xy0 = self.xy_prev
            else:
                self.load_0(xy0)
                xy0 = self.xy_prev
        elif type(xy0) == dict:
            with open('xy_0.json','w') as fobj:
                fobj.write(json.dumps(xy0))
            self.load_0('xy_0.json')
            xy0 = self.xy_prev            
        else:
            if xy0 == 0:
                xy0 = np.zeros(self.N_x+self.N_y)
            elif xy0 == 1:
                xy0 = np.ones(self.N_x+self.N_y)
            else:
                xy0 = xy0*np.ones(self.N_x+self.N_y)

        #xy = sopt.fsolve(self.ini_problem,xy0, jac=self.ini_dae_jacobian )

        
        if self.sopt_root_jac:
            sol = sopt.root(self.ini_problem, xy0, 
                            jac=self.ini_dae_jacobian, 
                            method=self.sopt_root_method, tol=self.initialization_tol)
        else:
            sol = sopt.root(self.ini_problem, xy0, method=self.sopt_root_method)

        self.initialization_ok = True
        if sol.success == False:
            print('initialization not found!')
            self.initialization_ok = False

            T = self.struct[0]['T'][:self.struct[0].it_store]
            X = self.struct[0]['X'][:self.struct[0].it_store,:]
            Y = self.struct[0]['Y'][:self.struct[0].it_store,:]
            Z = self.struct[0]['Z'][:self.struct[0].it_store,:]
            iters = self.struct[0]['iters'][:self.struct[0].it_store,:]

        if self.initialization_ok:
            xy = sol.x
            self.xy_prev = xy
            self.struct[0].x[:,0] = xy[0:self.N_x]
            self.struct[0].y_run[:,0] = xy[self.N_x:]

            ## y_ini to u_run
            for item in self.inputs_run_list:
                if item in self.y_ini_list:
                    self.struct[0][item] = self.struct[0].y_ini[self.y_ini_list.index(item)]

            ## u_ini to y_run
            for item in self.inputs_ini_list:
                if item in self.y_run_list:
                    self.struct[0].y_run[self.y_run_list.index(item)] = self.struct[0][item]


            #xy = sopt.fsolve(self.ini_problem,xy0, jac=self.ini_dae_jacobian )
            if self.sopt_root_jac:
                sol = sopt.root(self.run_problem, xy0, 
                                jac=self.run_dae_jacobian, 
                                method=self.sopt_root_method, tol=self.initialization_tol)
            else:
                sol = sopt.root(self.run_problem, xy0, method=self.sopt_root_method)

            if self.compile:
                # evaluate f and g
                run(0.0,self.struct,2)
                run(0.0,self.struct,3)                
    
                # evaluate run jacobians 
                run(0.0,self.struct,10)
                run(0.0,self.struct,11)                
                run(0.0,self.struct,12) 
                run(0.0,self.struct,14) 
                
            else:
                # evaluate f and g
                run.py_func(0.0,self.struct,2)
                run.py_func(0.0,self.struct,3)                
    
                # evaluate run jacobians 
                run.py_func(0.0,self.struct,10)
                run.py_func(0.0,self.struct,11)                
                run.py_func(0.0,self.struct,12) 
                run.py_func(0.0,self.struct,14)                 
                
             
            # post process result    
            T = self.struct[0]['T'][:self.struct[0].it_store]
            X = self.struct[0]['X'][:self.struct[0].it_store,:]
            Y = self.struct[0]['Y'][:self.struct[0].it_store,:]
            Z = self.struct[0]['Z'][:self.struct[0].it_store,:]
            iters = self.struct[0]['iters'][:self.struct[0].it_store,:]
        
            self.T = T
            self.X = X
            self.Y = Y
            self.Z = Z
            self.iters = iters
            
        return self.initialization_ok
    
    
    def get_value(self,name):
        if name in self.inputs_run_list:
            value = self.struct[0][name]
        if name in self.x_list:
            idx = self.x_list.index(name)
            value = self.struct[0].x[idx,0]
        if name in self.y_run_list:
            idy = self.y_run_list.index(name)
            value = self.struct[0].y_run[idy,0]
        if name in self.params_list:
            value = self.struct[0][name]
        if name in self.outputs_list:
            value = self.struct[0].h[self.outputs_list.index(name),0] 

        return value
    
    def get_values(self,name):
        if name in self.x_list:
            values = self.X[:,self.x_list.index(name)]
        if name in self.y_run_list:
            values = self.Y[:,self.y_run_list.index(name)]
        if name in self.outputs_list:
            values = self.Z[:,self.outputs_list.index(name)]
                        
        return values

    def get_mvalue(self,names):
        '''

        Parameters
        ----------
        names : list
            list of variables names to return each value.

        Returns
        -------
        mvalue : TYPE
            list of value of each variable.

        '''
        mvalue = []
        for name in names:
            mvalue += [self.get_value(name)]
                        
        return mvalue
    
    def set_value(self,name_,value):
        if name_ in self.inputs_run_list:
            self.struct[0][name_] = value
            return
        elif name_ in self.params_list:
            self.struct[0][name_] = value
            return
        elif name_ in self.inputs_ini_list:
            self.struct[0][name_] = value
            return 
        else:
            print(f'Input or parameter {name_} not found.')

    def set_values(self,dictionary):
        
        for item in dictionary:
            self.set_value(item,dictionary[item])
            
            
    def report_x(self,value_format='5.2f', decimals=2):
        for item in self.x_list:
            print(f'{item:5s} = {self.get_value(item):5.{decimals}f}')

    def report_y(self,value_format='5.2f', decimals=2):
        for item in self.y_run_list:
            print(f'{item:5s} = {self.get_value(item):5.{decimals}f}')
            
    def report_u(self,value_format='5.2f', decimals=2):
        for item in self.inputs_run_list:
            print(f'{item:5s} = {self.get_value(item):5.{decimals}f}')

    def report_z(self,value_format='5.2f', decimals=2):
        for item in self.outputs_list:
            print(f'{item:5s} = {self.get_value(item):5.{decimals}f}')

    def report_params(self,value_format='5.2f', decimals=2):
        for item in self.params_list:
            print(f'{item:5s} = {self.get_value(item):5.{decimals}f}')
            
    def get_x(self):
        return self.struct[0].x
    
    def ss(self):
        
        ssate(self.struct,self.xy_prev.reshape(len(self.xy_prev),1))
        
        ## y_ini to y_run
        self.struct[0].y_run = self.struct[0].y_ini
        
        ## y_ini to u_run
        for item in self.yini2urun:
            self.struct[0][item] = self.struct[0].y_ini[self.y_ini_list.index(item)]
                
        ## u_ini to y_run
        for item in self.uini2yrun:
            self.struct[0].y_run[self.y_run_list.index(item)] = self.struct[0][item]






# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 23:36:15 2020

@author: jmmau
"""

import numpy as np
import pandas as pd 
import numba
import scipy.optimize as sopt
class yoko:
    
    def __init__(self):
        
        self.names = ['error','i_a','v_a','i_b','v_b','i_c','v_c']
        self.t_offset = 0.0
        self.Dt_pll = 0.04
        
    def read(self,file_path):  
        
        
        with open(file_path,'r') as fobj:
            lines = fobj.readlines()

        self.meta_data = {}
        for it in range(14):
            parameter_value = lines[it].split(',')
            key = parameter_value[0].replace('"','').strip()
            value = parameter_value[1].replace('"','').replace('\n','').strip()
            #print(value)
            if value.isdecimal(): value = float(value)
            try: 
                value = float(value)
            except ValueError:
                pass

            self.meta_data.update({key:value})
        df_exp = pd.read_csv(file_path,skiprows=14,header=0,names=self.names,index_col=False)

        time_interval = self.meta_data['BlockSize']*self.meta_data['HResolution']
        df_exp['Time'] = np.linspace(self.t_offset,self.t_offset+time_interval,int(self.meta_data['BlockSize']))
        df_exp.drop('error',axis='columns', inplace=True)
        self.df  = df_exp

    def abc2pq(self,omega=2*np.pi*50,theta_0=0.0):

        times = self.df.Time.values 
        freq = 50.0
        omega = 2*np.pi*freq
        v_a = self.df.v_a.values
        v_b = self.df.v_b.values
        v_c = self.df.v_c.values
        i_a = self.df.i_a.values
        i_b = self.df.i_b.values
        i_c = self.df.i_c.values




        Dt = times[1]-times[0] 
        p = times*0.0
        q = times*0.0
        for it in range(len(times)):

            theta = Dt*it*omega + theta_0
            v_abc = np.array([[v_a[it]],[v_b[it]],[v_c[it]]])
            T_p = 2.0/3.0*np.array([[ np.cos(theta), np.cos(theta-2.0/3.0*np.pi), np.cos(theta+2.0/3.0*np.pi)],
                                    [-np.sin(theta),-np.sin(theta-2.0/3.0*np.pi),-np.sin(theta+2.0/3.0*np.pi)]])

            dq=T_p@v_abc;

            v_d = dq[0]
            v_q = dq[1]

            theta = Dt*it*omega + theta_0
            i_abc = np.array([[i_a[it]],[i_b[it]],[i_c[it]]])
            T_p = 2.0/3.0*np.array([[ np.cos(theta), np.cos(theta-2.0/3.0*np.pi), np.cos(theta+2.0/3.0*np.pi)],
                                    [-np.sin(theta),-np.sin(theta-2.0/3.0*np.pi),-np.sin(theta+2.0/3.0*np.pi)]])

            i_dq=T_p@i_abc;

            i_d = i_dq[0]
            i_q = i_dq[1]

            p[it] = 3/2*(v_d*i_d + v_q*i_q)
            q[it] = 3/2*(v_d*i_q - v_q*i_d)
            
        self.p = p
        self.q = q


    def pll(self):
        
        df_time = self.df['Time'].values
        times = np.linspace(df_time[0],df_time[-1]-self.Dt_pll,1000)
        theta = np.zeros(len(times))
        omega = np.zeros(len(times))
        it = 0
        a = np.max(self.df['v_a'].values)
        b = 2*np.pi*50
        c = 0.0
        d = 0.0
        for t in times:
            t_0 = t
            idx_1 = np.searchsorted(self.df['Time'],t_0)
            idx_2 = np.searchsorted(self.df['Time'],t_0 + self.Dt_pll)
            decimation = 10
            xdata = self.df['Time'][idx_1:idx_2:decimation].values
            ydata = self.df['v_a'][idx_1:idx_2:decimation].values
    
    
            popt, pcov = sopt.curve_fit(func, xdata, ydata, p0=[a,b,c,d])
            a,b,c,d = popt
            theta[it] = b*t+c
            omega[it] = b
            it += 1
    
        self.df['theta_pll'] = np.interp(df_time,times,theta)
        self.df['omega_pll'] = np.interp(df_time,times,omega)



@numba.njit(cache=True)
def func(x, a, b, c, d):
    y = np.copy(x)
    y = a * np.sin(b * x + c) + d
    return y

@numba.njit(cache=True)
def abc2dq(times,a,b,c,Theta):
    N = len(times)
    d = np.zeros((N,1))
    q = np.zeros((N,1))
    for it in range(len(times)):

        abc = np.array([[a[it]],[b[it]],[c[it]]])
        theta = Theta[it]
        T_p = 2.0/3.0*np.array([[ np.cos(theta), np.cos(theta-2.0/3.0*np.pi), np.cos(theta+2.0/3.0*np.pi)],
                                [-np.sin(theta),-np.sin(theta-2.0/3.0*np.pi),-np.sin(theta+2.0/3.0*np.pi)]])

        dq=T_p@abc;
        
        d[it] = dq[0]
        q[it] = dq[1]
    
    return d,q



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 24 09:57:25 2020

@author: jmmauricio
"""
import numpy as np


def get_v(syst,nodes_dict):
    
    for node in nodes_dict:
        v = syst.get_value(f'v_{node}')
        nodes_dict[node].update({'v':v})

def get_vp(syst,nodes_dict):
    
    for node in nodes_dict:
        v = syst.get_value(f'v_{node}')
        p = syst.get_value(f'p_{node}')
        nodes_dict[node].update({'v':v,'p':p})




def trains_update(t,trains,trips):
    '''
    trips = {
        'right':{'times':t_right,'positions':x_right,'powers':p_right},
        'left': {'times': t_left,'positions': x_left,'powers': p_left},
         }
    '''
    # train positions and powers updates
    train_positions = []
    train_powers = []
    train_idxs = []
    train_idts = []
    for train in trains:
        if t>train['t_ini']*60:  # the train started the trip
            trip = train['trip']
            t_trip = trips[trip]['times']
            if t<(train['t_ini']*60+t_trip[-1]):   # the train if still travelling 
                x_trip = trips[trip]['positions']
                p_trip = trips[trip]['powers']
                train_idt = np.argmin(np.array(t_trip)<(t-train['t_ini']*60)) 
                train_positions += [x_trip[train_idt]]
                train_powers    += [p_trip[train_idt]]
                train_idts  += [train_idt]   
    
    return train_positions,train_powers


def trains2params(sections,train_positions_list,train_powers_list,r_m):
    
    train_positions = np.array(train_positions_list) 
    train_powers = np.array(train_powers_list) 
    sort_index = np.argsort(train_positions) 

    train_positions_sorted =  train_positions[sort_index] 
    train_powers_sorted =  train_powers[sort_index] 

    for section in sections:
        section['N_trains'] = 0
        section['T_power'] = []
        section['T_pos'] = []
        
        
    # trains per section
    for train_position,train_power in zip(train_positions_sorted,train_powers_sorted):
        for section in sections:
            if section["pos_ini"] < train_position < section["pos_end"]:
                section['N_trains']+=1
                section['T_power']+=[train_power]
                section['T_pos']+=[train_position]    

    params_dict = {}
    for section in sections:
        nodes = section['nodes_i'] + section['nodes_v']
        N_sections = len(nodes)-1
        section['length'] =  section["pos_end"] - section["pos_ini"] 
        section['N_sections'] =  N_sections
        section['sub_length'] = []

        if section['N_trains'] == 0:
            for it in range(section['N_sections']):
                section['sub_length'] += [section['length']/section['N_sections']]

        if section['N_trains'] == 1:
            length = section['T_pos'][0] - section["pos_ini"]
            section['sub_length'] +=  [length]
            for it in range(1,section['N_sections']):
                section['sub_length'] += [(section['length']-length)/(section['N_sections']-1)]  

        if section['N_trains'] == 2:
            length_1 = section['T_pos'][0] - section["pos_ini"]
            section['sub_length'] +=  [length_1]
            length_2 = section['T_pos'][1] - section['T_pos'][0]
            section['sub_length'] +=  [length_2]
            for it in range(2,section['N_sections']):
                section['sub_length'] += [(section['length']-(length_1+length_2))/(section['N_sections']-2)]     

        if section['N_trains'] == 3:
            length_1 = section['T_pos'][0] - section["pos_ini"]
            section['sub_length'] +=  [length_1]
            length_2 = section['T_pos'][1] - section['T_pos'][0]
            section['sub_length'] +=  [length_2]
            length_3 = section['T_pos'][2] - section['T_pos'][1]
            section['sub_length'] +=  [length_3]
            for it in range(3,section['N_sections']):
                section['sub_length'] += [(section['length']-(length_1+length_2+length_3))/(section['N_sections']-3)] 

        if section['N_trains'] == 4:
            length_1 = section['T_pos'][0] - section["pos_ini"]
            section['sub_length'] +=  [length_1]
            length_2 = section['T_pos'][1] - section['T_pos'][0]
            section['sub_length'] +=  [length_2]
            length_3 = section['T_pos'][2] - section['T_pos'][1]
            section['sub_length'] +=  [length_3]
            length_4 = section['T_pos'][3] - section['T_pos'][2]
            section['sub_length'] +=  [length_4]
            for it in range(3,section['N_sections']):
                section['sub_length'] += [(section['length']-(length_1+length_2+length_3+length_4))/(section['N_sections']-4)] 



    # impedance between nodes
    for section in sections:
        nodes = sorted(section['nodes_i']+section['nodes_v'])
        for i_sub in range(len(nodes)-1):           
            j = nodes[i_sub]
            k = nodes[i_sub+1]
            params_dict.update({f'R_{j}{k}':r_m*section['sub_length'][i_sub]})

    # train powers to nodes_i
    i_sec = 0
    for section in sections:
        nodes_i = section['nodes_i']
        T_powers = section['T_power']
        for it in range(section['N_tnodes']):           
            params_dict.update({f'p_{nodes_i[it]}':0.0})
        for it in range(section['N_trains']):       
            if i_sec == 0: params_dict.update({f'p_{nodes_i[it+1]}':-T_powers[it]})
            if i_sec >  0: params_dict.update({f'p_{nodes_i[it]}':-T_powers[it]})
        i_sec += 1

    nodes_dict = {}
    abs_length = 0.0
    for section in sections:
        nodes_i = section['nodes_i']
        nodes_v = section['nodes_v']
        section_nodes = sorted(nodes_i + nodes_v)
        for length,node in zip(section['sub_length'],section_nodes[:-1]):
            nodes_dict.update({node:{'pos':abs_length}})
            abs_length += length
    nodes_dict.update({section_nodes[-1]:{'pos':abs_length}})


    return params_dict,nodes_dict



# -*- coding: utf-8 -*-
"""
Created on Thu May 28 19:54:52 2020

@author: jmmau
"""

        #perturbations(t,struct) 
        
        # if solver == 1: 
        #     # forward euler solver  
        #     run(t,struct, 2)  
        #     struct[i].x[:] += Dt*struct[i].f  
 
        # if solver == 2: 
            
        #     # bacward euler solver
        #     x_0 = np.copy(struct[i].x[:]) 
        #     for j in range(struct[i].imax): 
        #         run(t,struct, 2) 
        #         run(t,struct, 3) 
        #         run(t,struct, 10)  
        #         phi =  x_0 + Dt*struct[i].f - struct[i].x 
        #         Dx = np.linalg.solve(-(Dt*struct[i].Fx - np.eye(N_x)), phi) 
        #         struct[i].x[:] += Dx[:] 
        #         if np.max(np.abs(Dx)) < struct[i].itol: break 
        #     print(struct[i].f)
 
        # if solver == 3: 
        #     # trapezoidal solver
        #     run(t,struct, 2) 
        #     f_0 = np.copy(struct[i].f[:]) 
        #     x_0 = np.copy(struct[i].x[:]) 
        #     for j in range(struct[i].imax): 
        #         run(t,struct, 10)  
        #         phi =  x_0 + 0.5*Dt*(f_0 + struct[i].f) - struct[i].x 
        #         Dx = np.linalg.solve(-(0.5*Dt*struct[i].Fx - np.eye(N_x)), phi) 
        #         struct[i].x[:] += Dx[:] 
        #         run(t,struct, 2) 
        #         if np.max(np.abs(Dx)) < struct[i].itol: break 

        # if solver == 4: # Teapezoidal DAE as in Milano's book

        #     run(t,struct, 2) 
        #     run(t,struct, 3) 

        #     x = np.copy(struct[i].x[:]) 
        #     y = np.copy(struct[i].y[:]) 
        #     f = np.copy(struct[i].f[:]) 
        #     g = np.copy(struct[i].g[:]) 
            
        #     for iter in range(struct[i].imax):
        #         run(t,struct, 2) 
        #         run(t,struct, 3) 
        #         run(t,struct,10) 
        #         run(t,struct,11) 
                
        #         x_i = struct[i].x[:] 
        #         y_i = struct[i].y[:]  
        #         f_i = struct[i].f[:] 
        #         g_i = struct[i].g[:]                 
        #         F_x_i = struct[i].Fx[:,:]
        #         F_y_i = struct[i].Fy[:,:] 
        #         G_x_i = struct[i].Gx[:,:] 
        #         G_y_i = struct[i].Gy[:,:]                

        #         A_c_i = np.vstack((np.hstack((eye-0.5*Dt*F_x_i, -0.5*Dt*F_y_i)),
        #                            np.hstack((G_x_i,         G_y_i))))
                     
        #         f_n_i = x_i - x - 0.5*Dt*(f_i+f) 
        #         # print(t,iter,g_i)
        #         Dxy_i = np.linalg.solve(-A_c_i,np.vstack((f_n_i,g_i))) 
                
        #         x_i = x_i + Dxy_i[0:N_x]
        #         y_i = y_i + Dxy_i[N_x:(N_x+N_y)]

        #         struct[i].x[:] = x_i
        #         struct[i].y[:] = y_i

        #         if np.max(np.abs(Dxy_i[:,0]))<struct[i].itol:
                    
        #             break
                
        #         # if iter>struct[i].imax-2:
                    
        #         #     print('Convergence problem')

        #     struct[i].x[:] = x_i
        #     struct[i].y[:] = y_i